from enum import Enum

from List.configDict import locationType, itemType
from List.inventory import magic, keyblade, ability, misc, report
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, item_bonus, hybrid_bonus, stat_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    Passage = "Passage"
    PassageChests = "Passage Chests"
    InnerChamber = "Inner Chamber"
    InnerChamberChests = "Inner Chamber Chests"
    CerberusBonus = "Cerberus Bonus"
    ColiseumMapPopup = "Coliseum Map Popup"
    UrnsBonus = "Urns Bonus"
    UnderworldEntrance = "Underworld Entrance"
    UnderworldEntranceChests = "Underworld Entrance Chests"
    CavernsEntrance = "Caverns Entrance"
    CavernsEntranceChests = "Caverns Entrance Chests"
    LostRoad = "Lost Road"
    LostRoadChests = "Lost Road Chests"
    Atrium = "Atrium"
    AtriumChests = "Atrium Chests"
    DemyxOlympusColiseum = "Demyx (OC)"
    TheLock = "The Lock"
    TheLockChests = "The Lock Chests"
    PeteOlympusColiseum = "Pete (OC)"
    Hydra = "Hydra"
    AuronsStatue = "Auron Statue"
    Hades = "Hades"
    PainAndPanicCup = "Pain and Panic Cup"
    CerberusCup = "Cerberus Cup"
    TitanCup = "Titan Cup"
    GoddessOfFateCup = "Goddess of Fate Cup"
    ParadoxCups = "Paradox Cups"
    Zexion = "AS Zexion"
    DataZexion = "Data Zexion"


class CheckLocation(str, Enum):
    PassageMythrilShard = "Passage Mythril Shard"
    PassageMythrilStone = "Passage Mythril Stone"
    PassageEther = "Passage Ether"
    PassageApBoost = "Passage AP Boost"
    PassageHiPotion = "Passage Hi-Potion"
    InnerChamberUnderworldMap = "Inner Chamber Underworld Map"
    InnerChamberMythrilShard = "Inner Chamber Mythril Shard"
    Cerberus = "Cerberus"
    ColiseumMap = "Coliseum Map"
    Urns = "Urns"
    UnderworldEntrancePowerBoost = "Underworld Entrance Power Boost"
    CavernsEntranceLucidShard = "Caverns Entrance Lucid Shard"
    CavernsEntranceApBoost = "Caverns Entrance AP Boost"
    CavernsEntranceMythrilShard = "Caverns Entrance Mythril Shard"
    LostRoadBrightShard = "The Lost Road Bright Shard"
    LostRoadEther = "The Lost Road Ether"
    LostRoadMythrilShard = "The Lost Road Mythril Shard"
    LostRoadMythrilStone = "The Lost Road Mythril Stone"
    AtriumLucidStone = "Atrium Lucid Stone"
    AtriumApBoost = "Atrium AP Boost"
    DemyxOlympusColiseum = "Demyx OC"
    DemyxOcSecretAnsemReport5 = "Secret Ansem Report 5"
    OlympusStone = "Olympus Stone"
    LockCavernsMap = "The Lock Caverns Map"
    LockMythrilShard = "The Lock Mythril Shard"
    LockApBoost = "The Lock AP Boost"
    PeteOlympusColiseum = "Pete (OC)"
    Hydra = "Hydra"
    HerosCrest = "Hero's Crest"
    AuronssStatue = "Auron's Statue"
    Hades = "Hades"
    GuardianSoul = "Guardian Soul"
    PainPanicCupProtectBelt = "Protect Belt (Pain and Panic Cup)"
    PainPanicCupSerenityGem = "Serenity Gem (Pain and Panic Cup)"
    CerberusCupRisingDragon = "Rising Dragon (Cerberus Cup)"
    CerberusCupSerenityCrystal = "Serenity Crystal (Cerberus Cup)"
    TitanCupGenjiShield = "Genji Shield (Titan Cup)"
    TitanCupSkillfulRing = "Skillful Ring (Titan Cup)"
    GoddessOfFateCupFatalCrest = "Fatal Crest (Goddess of Fate Cup)"
    GoddessOfFateCupOrichalcumPlus = "Orichalcum+ (Goddess of Fate Cup)"
    ParadoxCupsHadesCupTrophy = "Hades Cup Trophy (Paradox Cups)"
    ZexionBonus = "Zexion Bonus"
    ZexionBookOfShadows = "Zexion (AS) Book of Shadows"
    DataZexionLostIllusion = "Zexion (Data) Lost Illusion"

class OCLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_oc_keyblade(inv)
        self.logic[NodeId.Passage][NodeId.PassageChests] = keyblade_lambda
        self.logic[NodeId.InnerChamber][NodeId.InnerChamberChests] = keyblade_lambda
        self.logic[NodeId.UnderworldEntrance][NodeId.UnderworldEntranceChests] = keyblade_lambda
        self.logic[NodeId.CavernsEntrance][NodeId.CavernsEntranceChests] = keyblade_lambda
        self.logic[NodeId.LostRoad][NodeId.LostRoadChests] = keyblade_lambda
        self.logic[NodeId.Atrium][NodeId.AtriumChests] = keyblade_lambda
        self.logic[NodeId.TheLock][NodeId.TheLockChests] = keyblade_lambda
        pain_and_panic_check = ItemPlacementHelpers.dc2_check
        cerberus_cup_check = lambda inv: ItemPlacementHelpers.ht1_check(inv) and ItemPlacementHelpers.ag1_check(inv) and ItemPlacementHelpers.pl1_check(inv)
        goddess_of_fate_check = lambda inv: cerberus_cup_check(inv) and pain_and_panic_check(inv) and ItemPlacementHelpers.twtnw_post_saix_check(inv)
        if not reverse_rando:
            self.logic[START_NODE][NodeId.Passage] = ItemPlacementHelpers.oc1_check
            self.logic[NodeId.Hydra][NodeId.AuronsStatue] = ItemPlacementHelpers.oc2_check
            self.logic[NodeId.Hydra][NodeId.PainAndPanicCup] = pain_and_panic_check
            self.logic[NodeId.Hydra][NodeId.CerberusCup] = cerberus_cup_check
            self.logic[NodeId.Hades][NodeId.GoddessOfFateCup] = goddess_of_fate_check
            self.logic[NodeId.Hades][NodeId.ParadoxCups] = lambda inv: ItemPlacementHelpers.need_forms(inv) and ItemPlacementHelpers.need_summons(inv)
        else:
            self.logic[START_NODE][NodeId.UnderworldEntrance] = ItemPlacementHelpers.oc1_check
            self.logic[NodeId.Zexion][NodeId.CerberusBonus] = ItemPlacementHelpers.oc2_check
            self.logic[NodeId.Hades][NodeId.PainAndPanicCup] = pain_and_panic_check
            self.logic[NodeId.Hades][NodeId.CerberusCup] = cerberus_cup_check
            self.logic[NodeId.Hydra][NodeId.GoddessOfFateCup] = goddess_of_fate_check
            self.logic[NodeId.Hydra][NodeId.ParadoxCups] = lambda inv: ItemPlacementHelpers.need_forms(inv) and ItemPlacementHelpers.need_summons(inv)

def make_graph(graph: LocationGraphBuilder):
    oc = locationType.OC
    cups = locationType.OCCups
    oc_logic = OCLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(oc_logic)

    passage_chests = graph.add_location(NodeId.PassageChests, [
        chest(7, CheckLocation.PassageMythrilShard, oc),
        chest(8, CheckLocation.PassageMythrilStone, oc),
        chest(144, CheckLocation.PassageEther, oc),
        chest(145, CheckLocation.PassageApBoost, oc),
        chest(146, CheckLocation.PassageHiPotion, oc),
    ])
    passage = graph.add_location(NodeId.Passage, [])
    inner_chamber_chests = graph.add_location(NodeId.InnerChamberChests, [
        chest(2, CheckLocation.InnerChamberUnderworldMap, oc),
        chest(243, CheckLocation.InnerChamberMythrilShard, oc),
    ])
    inner_chamber = graph.add_location(NodeId.InnerChamber, [])
    cerberus_bonus = graph.add_location(NodeId.CerberusBonus, [
        item_bonus(5, CheckLocation.Cerberus, oc, vanilla=ability.Counterguard),
    ])
    coliseum_map_popup = graph.add_location(NodeId.ColiseumMapPopup, [
        popup(338, CheckLocation.ColiseumMap, oc),
    ])
    urns_bonus = graph.add_location(NodeId.UrnsBonus, [
        item_bonus(57, CheckLocation.Urns, oc, vanilla=ability.AerialDive),
    ])
    underworld_entrance_chests = graph.add_location(NodeId.UnderworldEntranceChests, [
        chest(242, CheckLocation.UnderworldEntrancePowerBoost, oc),
    ])
    underworld_entrance = graph.add_location(NodeId.UnderworldEntrance, [])
    caverns_entrance_chests = graph.add_location(NodeId.CavernsEntranceChests, [
        chest(3, CheckLocation.CavernsEntranceLucidShard, oc),
        chest(11, CheckLocation.CavernsEntranceApBoost, oc),
        chest(504, CheckLocation.CavernsEntranceMythrilShard, oc),
    ])
    caverns_entrance = graph.add_location(NodeId.CavernsEntrance, [])
    lost_road_chests = graph.add_location(NodeId.LostRoadChests, [
        chest(9, CheckLocation.LostRoadBrightShard, oc),
        chest(10, CheckLocation.LostRoadEther, oc),
        chest(148, CheckLocation.LostRoadMythrilShard, oc),
        chest(149, CheckLocation.LostRoadMythrilStone, oc),
    ])
    lost_road = graph.add_location(NodeId.LostRoad, [])
    atrium_chests = graph.add_location(NodeId.AtriumChests, [
        chest(150, CheckLocation.AtriumLucidStone, oc),
        chest(151, CheckLocation.AtriumApBoost, oc),
    ])
    atrium = graph.add_location(NodeId.Atrium, [])
    demyx = graph.add_location(NodeId.DemyxOlympusColiseum, [
        stat_bonus(58, CheckLocation.DemyxOlympusColiseum, oc),
        popup(529, CheckLocation.DemyxOcSecretAnsemReport5, oc, vanilla=report.AnsemReport5),
        popup(293, CheckLocation.OlympusStone, oc, vanilla=misc.OlympusStone),
    ])
    lock_chests = graph.add_location(NodeId.TheLockChests, [
        chest(244, CheckLocation.LockCavernsMap, oc),
        chest(5, CheckLocation.LockMythrilShard, oc),
        chest(142, CheckLocation.LockApBoost, oc),
    ])
    lock = graph.add_location(NodeId.TheLock, [])
    pete = graph.add_location(NodeId.PeteOlympusColiseum, [
        item_bonus(6, CheckLocation.PeteOlympusColiseum, oc, vanilla=ability.TrinityLimit),
    ])
    hydra = graph.add_location(NodeId.Hydra, [
        hybrid_bonus(7, CheckLocation.Hydra, oc, vanilla=magic.Thunder),
        popup(260, CheckLocation.HerosCrest, oc, vanilla=keyblade.HerosCrest),
    ])
    aurons_statue = graph.add_location(NodeId.AuronsStatue, [
        popup(295, CheckLocation.AuronssStatue, oc, vanilla=misc.AuronsStatue),
    ])
    hades = graph.add_location(NodeId.Hades, [
        hybrid_bonus(8, CheckLocation.Hades, oc, vanilla=ability.MagnetBurst),
        popup(272, CheckLocation.GuardianSoul, oc, vanilla=keyblade.GuardianSoul),
    ])
    pain_panic_cup = graph.add_location(NodeId.PainAndPanicCup, [
        popup(513, CheckLocation.PainPanicCupProtectBelt, [oc, cups], invalid_checks=[itemType.TROPHY]),
        popup(540, CheckLocation.PainPanicCupSerenityGem, [oc, cups], invalid_checks=[itemType.TROPHY]),
    ])
    cerberus_cup = graph.add_location(NodeId.CerberusCup, [
        popup(515, CheckLocation.CerberusCupRisingDragon, [oc, cups], invalid_checks=[itemType.TROPHY]),
        popup(542, CheckLocation.CerberusCupSerenityCrystal, [oc, cups], invalid_checks=[itemType.TROPHY]),
    ])
    titan_cup = graph.add_location(NodeId.TitanCup, [
        popup(514, CheckLocation.TitanCupGenjiShield, [oc, cups], invalid_checks=[itemType.TROPHY]),
        popup(541, CheckLocation.TitanCupSkillfulRing, [oc, cups], invalid_checks=[itemType.TROPHY]),
    ])
    goddess_of_fate_cup = graph.add_location(NodeId.GoddessOfFateCup, [
        popup(516, CheckLocation.GoddessOfFateCupFatalCrest, [oc, cups], invalid_checks=[itemType.TROPHY],
              vanilla=keyblade.FatalCrest),
        popup(517, CheckLocation.GoddessOfFateCupOrichalcumPlus, [oc, cups], invalid_checks=[itemType.TROPHY]),
    ])
    paradox_cups = graph.add_location(NodeId.ParadoxCups, [
        popup(518, CheckLocation.ParadoxCupsHadesCupTrophy, [oc, locationType.OCParadoxCup],
              invalid_checks=[itemType.TROPHY, itemType.FORM, itemType.SUMMON], vanilla=misc.HadesCupTrophy),
    ])
    zexion = graph.add_location(NodeId.Zexion, [
        stat_bonus(66, CheckLocation.ZexionBonus, [oc, locationType.AS]),
        popup(546, CheckLocation.ZexionBookOfShadows, [oc, locationType.AS]),
    ])
    data_zexion = graph.add_location(NodeId.DataZexion, [
        popup(551, CheckLocation.DataZexionLostIllusion, [oc, locationType.DataOrg]),
    ])

    graph.register_superboss(data_zexion)
    
    graph.add_edge(passage, passage_chests)
    graph.add_edge(inner_chamber, inner_chamber_chests)
    graph.add_edge(underworld_entrance, underworld_entrance_chests)
    graph.add_edge(caverns_entrance, caverns_entrance_chests)
    graph.add_edge(lost_road, lost_road_chests)
    graph.add_edge(atrium, atrium_chests)
    graph.add_edge(lock, lock_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, passage)
        graph.add_edge(passage, inner_chamber)
        graph.add_edge(inner_chamber, cerberus_bonus, RequirementEdge(battle=True))
        graph.add_edge(cerberus_bonus, coliseum_map_popup)
        graph.add_edge(coliseum_map_popup, urns_bonus)
        graph.add_edge(urns_bonus, underworld_entrance)
        graph.add_edge(underworld_entrance, caverns_entrance)
        graph.add_edge(caverns_entrance, lost_road)
        graph.add_edge(lost_road, atrium)
        graph.add_edge(atrium, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, lock)
        graph.add_edge(lock, pete, RequirementEdge(battle=True))
        graph.add_edge(pete, hydra, RequirementEdge(battle=True))
        graph.add_edge(hydra, aurons_statue, RequirementEdge(battle=True))
        graph.add_edge(aurons_statue, hades, RequirementEdge(battle=True))

        graph.add_edge(hydra, pain_panic_cup, RequirementEdge(battle=True))
        graph.add_edge(hydra, cerberus_cup, RequirementEdge(battle=True))
        graph.add_edge(hades, titan_cup, RequirementEdge(battle=True))
        graph.add_edge(hades, goddess_of_fate_cup, RequirementEdge(battle=True))
        graph.add_edge(hades, paradox_cups)
        graph.add_edge(hades, zexion, RequirementEdge(battle=True))
        graph.add_edge(zexion, data_zexion)
        graph.register_first_boss(hydra)
        graph.register_last_story_boss(hades)
        graph.register_superboss(zexion)
    else:
        graph.add_edge(START_NODE, underworld_entrance)
        graph.add_edge(underworld_entrance, passage)
        graph.add_edge(passage, inner_chamber)
        graph.add_edge(inner_chamber, aurons_statue, RequirementEdge(battle=True))
        graph.add_edge(aurons_statue, hades, RequirementEdge(battle=True))
        graph.add_edge(hades, zexion, RequirementEdge(battle=True))
        graph.add_edge(zexion, cerberus_bonus, RequirementEdge(battle=True))
        graph.add_edge(cerberus_bonus, coliseum_map_popup)
        graph.add_edge(coliseum_map_popup, urns_bonus)
        graph.add_edge(urns_bonus, caverns_entrance)
        graph.add_edge(caverns_entrance, lost_road)
        graph.add_edge(lost_road, atrium)
        graph.add_edge(atrium, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, lock)
        graph.add_edge(lock, pete, RequirementEdge(battle=True))
        graph.add_edge(pete, hydra, RequirementEdge(battle=True))
        graph.add_edge(hades, pain_panic_cup, RequirementEdge(battle=True))
        graph.add_edge(hades, cerberus_cup, RequirementEdge(battle=True))
        graph.add_edge(hydra, titan_cup, RequirementEdge(battle=True))
        graph.add_edge(hydra, goddess_of_fate_cup, RequirementEdge(battle=True))
        graph.add_edge(hydra, paradox_cups)
        graph.add_edge(hydra, data_zexion, RequirementEdge(battle=True))
        graph.register_first_boss(hades)
        graph.register_last_story_boss(hydra)



import itertools
from typing import Optional, Iterable

from Class.itemClass import KH2Item
from List.configDict import itemRarity
from List.inventory import ability, accessory, armor, bonus, consumable, form, growth, keyblade, magic, \
    misc, map, proof, recipe, report, shield, staff, storyunlock, summon, synth
from List.inventory.ability import DonaldAbilities, GoofyAbilities, Ability
from List.inventory.keyblade import Keyblade
from List.location import weaponslot


class Items:

    @staticmethod
    def sora_lookup_table() -> dict[int, KH2Item]:
        """ Returns an _attempt_ at a comprehensive lookup table of item ID to item. May not include everything. """
        full_list = Items.getItemList()
        full_list.extend(Items.getStatItems())
        full_list.extend(Items.getDummyFormItems())
        full_list.append(Items.getPromiseCharm())
        full_list.append(Items.getTT1Jailbreak())
        full_list.append(Items.objectiveReportItem())
        full_list.extend(Items.getSupportAbilityList())
        full_list.extend(Items.getKeybladeAbilityList())
        full_list.extend(Items.getActionAbilityList())
        full_list.extend(Items.getLevelAbilityList())
        full_list.extend(Items.getJunkList(betterJunk=False))

        result = {}
        for item in full_list:
            item_id = item.Id
            if item_id not in result:
                result[item_id] = item
        return result

    @staticmethod
    def getItemList(story_unlocking_rarity: itemRarity = itemRarity.UNCOMMON) -> list[KH2Item]:
        return [
            KH2Item(proof.ProofOfConnection, itemRarity.MYTHIC),
            KH2Item(proof.ProofOfNonexistence, itemRarity.MYTHIC),
            KH2Item(proof.ProofOfPeace, itemRarity.MYTHIC),

            KH2Item(storyunlock.BattlefieldsOfWar, story_unlocking_rarity),
            KH2Item(storyunlock.BattlefieldsOfWar, story_unlocking_rarity),
            KH2Item(storyunlock.SwordOfTheAncestor, story_unlocking_rarity),
            KH2Item(storyunlock.SwordOfTheAncestor, story_unlocking_rarity),
            KH2Item(storyunlock.BeastsClaw, story_unlocking_rarity),
            KH2Item(storyunlock.BeastsClaw, story_unlocking_rarity),
            KH2Item(storyunlock.BoneFist, story_unlocking_rarity),
            KH2Item(storyunlock.BoneFist, story_unlocking_rarity),
            KH2Item(storyunlock.ProudFang, story_unlocking_rarity),
            KH2Item(storyunlock.ProudFang, story_unlocking_rarity),
            KH2Item(storyunlock.SkillAndCrossbones, story_unlocking_rarity),
            KH2Item(storyunlock.SkillAndCrossbones, story_unlocking_rarity),
            KH2Item(storyunlock.Scimitar, story_unlocking_rarity),
            KH2Item(storyunlock.Scimitar, story_unlocking_rarity),
            KH2Item(storyunlock.WayToTheDawn, story_unlocking_rarity),
            KH2Item(storyunlock.WayToTheDawn, story_unlocking_rarity),
            KH2Item(storyunlock.IdentityDisk, story_unlocking_rarity),
            KH2Item(storyunlock.IdentityDisk, story_unlocking_rarity),
            # KH2Item(storyunlock.TournamentPoster, story_unlocking_rarity),
            KH2Item(storyunlock.MembershipCard, story_unlocking_rarity),
            KH2Item(storyunlock.MembershipCard, story_unlocking_rarity),
            KH2Item(storyunlock.IceCream, story_unlocking_rarity),
            KH2Item(storyunlock.IceCream, story_unlocking_rarity),
            KH2Item(storyunlock.IceCream, story_unlocking_rarity),
            KH2Item(storyunlock.RoyalSummons, story_unlocking_rarity),
            KH2Item(storyunlock.RoyalSummons, story_unlocking_rarity),
            KH2Item(storyunlock.NaminesSketches, story_unlocking_rarity),

            KH2Item(report.AnsemReport1, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport2, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport3, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport4, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport5, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport6, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport7, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport8, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport9, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport10, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport11, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport12, itemRarity.UNCOMMON),
            KH2Item(report.AnsemReport13, itemRarity.UNCOMMON),

            KH2Item(magic.Fire, itemRarity.MYTHIC),
            KH2Item(magic.Fire, itemRarity.MYTHIC),
            KH2Item(magic.Fire, itemRarity.MYTHIC),

            KH2Item(magic.Blizzard, itemRarity.MYTHIC),
            KH2Item(magic.Blizzard, itemRarity.MYTHIC),
            KH2Item(magic.Blizzard, itemRarity.MYTHIC),

            KH2Item(magic.Thunder, itemRarity.MYTHIC),
            KH2Item(magic.Thunder, itemRarity.MYTHIC),
            KH2Item(magic.Thunder, itemRarity.MYTHIC),

            KH2Item(magic.Cure, itemRarity.MYTHIC),
            KH2Item(magic.Cure, itemRarity.MYTHIC),
            KH2Item(magic.Cure, itemRarity.MYTHIC),

            KH2Item(magic.Magnet, itemRarity.MYTHIC),
            KH2Item(magic.Magnet, itemRarity.MYTHIC),
            KH2Item(magic.Magnet, itemRarity.MYTHIC),

            KH2Item(magic.Reflect, itemRarity.MYTHIC),
            KH2Item(magic.Reflect, itemRarity.MYTHIC),
            KH2Item(magic.Reflect, itemRarity.MYTHIC),

            KH2Item(growth.HighJump1),
            KH2Item(growth.HighJump2, itemRarity.UNCOMMON),
            KH2Item(growth.HighJump3, itemRarity.RARE),
            KH2Item(growth.HighJumpMax, itemRarity.RARE),

            KH2Item(growth.QuickRun1),
            KH2Item(growth.QuickRun2, itemRarity.UNCOMMON),
            KH2Item(growth.QuickRun3, itemRarity.RARE),
            KH2Item(growth.QuickRunMax, itemRarity.RARE),

            KH2Item(growth.AerialDodge1),
            KH2Item(growth.AerialDodge2, itemRarity.UNCOMMON),
            KH2Item(growth.AerialDodge3, itemRarity.RARE),
            KH2Item(growth.AerialDodgeMax, itemRarity.RARE),

            KH2Item(growth.Glide1),
            KH2Item(growth.Glide2, itemRarity.UNCOMMON),
            KH2Item(growth.Glide3, itemRarity.RARE),
            KH2Item(growth.GlideMax, itemRarity.RARE),

            KH2Item(growth.DodgeRoll1),
            KH2Item(growth.DodgeRoll2, itemRarity.UNCOMMON),
            KH2Item(growth.DodgeRoll3, itemRarity.RARE),
            KH2Item(growth.DodgeRollMax, itemRarity.RARE),

            KH2Item(form.ValorForm, itemRarity.MYTHIC),
            KH2Item(form.WisdomForm, itemRarity.MYTHIC),
            KH2Item(form.FinalForm, itemRarity.MYTHIC),
            KH2Item(form.MasterForm, itemRarity.MYTHIC),
            KH2Item(form.LimitForm, itemRarity.MYTHIC),
            KH2Item(form.AntiForm, itemRarity.RARE),

            KH2Item(misc.TornPages, itemRarity.MYTHIC),
            KH2Item(misc.TornPages, itemRarity.MYTHIC),
            KH2Item(misc.TornPages, itemRarity.MYTHIC),
            KH2Item(misc.TornPages, itemRarity.MYTHIC),
            KH2Item(misc.TornPages, itemRarity.MYTHIC),

            KH2Item(summon.LampCharm, itemRarity.MYTHIC),
            KH2Item(summon.FeatherCharm, itemRarity.MYTHIC),
            KH2Item(summon.UkuleleCharm, itemRarity.MYTHIC),
            KH2Item(summon.BaseballCharm, itemRarity.MYTHIC),

            KH2Item(keyblade.Oathkeeper),
            KH2Item(keyblade.Oblivion),
            KH2Item(keyblade.StarSeeker),
            KH2Item(keyblade.HiddenDragon),
            KH2Item(keyblade.HerosCrest),
            KH2Item(keyblade.Monochrome),
            KH2Item(keyblade.FollowTheWind),
            KH2Item(keyblade.CircleOfLife),
            KH2Item(keyblade.PhotonDebugger),
            KH2Item(keyblade.GullWing),
            KH2Item(keyblade.RumblingRose),
            KH2Item(keyblade.GuardianSoul),
            KH2Item(keyblade.WishingLamp),
            KH2Item(keyblade.DecisivePumpkin),
            KH2Item(keyblade.SleepingLion),
            KH2Item(keyblade.SweetMemories),
            KH2Item(keyblade.MysteriousAbyss),
            KH2Item(keyblade.TwoBecomeOne),
            KH2Item(keyblade.FatalCrest),
            KH2Item(keyblade.BondOfFlame),
            KH2Item(keyblade.Fenrir),
            KH2Item(keyblade.UltimaWeapon),
            KH2Item(keyblade.WinnersProof),
            KH2Item(keyblade.Pureblood),

            KH2Item(staff.CenturionPlus, itemRarity.UNCOMMON),
            KH2Item(staff.MeteorStaff, itemRarity.UNCOMMON),
            KH2Item(staff.NobodyLance, itemRarity.UNCOMMON),
            KH2Item(staff.PreciousMushroom, itemRarity.UNCOMMON),
            KH2Item(staff.PreciousMushroomPlus, itemRarity.UNCOMMON),
            KH2Item(staff.PremiumMushroom, itemRarity.UNCOMMON),
            KH2Item(staff.RisingDragon, itemRarity.UNCOMMON),
            KH2Item(staff.SaveTheQueenPlus, itemRarity.UNCOMMON),
            KH2Item(staff.ShamansRelic, itemRarity.UNCOMMON),

            KH2Item(shield.AkashicRecord, itemRarity.UNCOMMON),
            KH2Item(shield.FrozenPridePlus, itemRarity.UNCOMMON),
            KH2Item(shield.GenjiShield, itemRarity.UNCOMMON),
            KH2Item(shield.MajesticMushroom, itemRarity.UNCOMMON),
            KH2Item(shield.MajesticMushroomPlus, itemRarity.UNCOMMON),
            KH2Item(shield.NobodyGuard, itemRarity.UNCOMMON),
            KH2Item(shield.OgreShield, itemRarity.UNCOMMON),
            KH2Item(shield.SaveTheKingPlus, itemRarity.UNCOMMON),
            KH2Item(shield.UltimateMushroom, itemRarity.UNCOMMON),

            KH2Item(accessory.AbilityRing),
            KH2Item(accessory.EngineersRing),
            KH2Item(accessory.TechniciansRing),
            KH2Item(accessory.SkillRing),
            KH2Item(accessory.SkillfulRing),
            KH2Item(accessory.ExpertsRing),
            KH2Item(accessory.MastersRing),
            KH2Item(accessory.CosmicRing),
            KH2Item(accessory.ExecutivesRing),
            KH2Item(accessory.SardonyxRing),
            KH2Item(accessory.TourmalineRing),
            KH2Item(accessory.AquamarineRing),
            KH2Item(accessory.GarnetRing),
            KH2Item(accessory.DiamondRing),
            KH2Item(accessory.SilverRing),
            KH2Item(accessory.GoldRing),
            KH2Item(accessory.PlatinumRing),
            KH2Item(accessory.MythrilRing),
            KH2Item(accessory.OrichalcumRing),
            KH2Item(accessory.SoldierEarring),
            KH2Item(accessory.FencerEarring, itemRarity.UNCOMMON),
            KH2Item(accessory.MageEarring),
            KH2Item(accessory.SlayerEarring, itemRarity.UNCOMMON),
            KH2Item(accessory.Medal),
            KH2Item(accessory.MoonAmulet, itemRarity.RARE),
            KH2Item(accessory.StarCharm, itemRarity.RARE),
            KH2Item(accessory.CosmicArts, itemRarity.RARE),
            KH2Item(accessory.ShadowArchive, itemRarity.UNCOMMON),
            KH2Item(accessory.ShadowArchivePlus, itemRarity.RARE),
            KH2Item(accessory.FullBloom, itemRarity.UNCOMMON),
            KH2Item(accessory.FullBloomPlus, itemRarity.RARE),
            KH2Item(accessory.DrawRing, itemRarity.UNCOMMON),
            KH2Item(accessory.LuckyRing),

            KH2Item(armor.ElvenBandanna),
            KH2Item(armor.DivineBandanna),
            KH2Item(armor.ProtectBelt),
            KH2Item(armor.GaiaBelt),
            KH2Item(armor.PowerBand),
            KH2Item(armor.BusterBand),
            KH2Item(armor.CosmicBelt),
            KH2Item(armor.FireBangle),
            KH2Item(armor.FiraBangle),
            KH2Item(armor.FiragaBangle),
            KH2Item(armor.FiragunBangle),
            KH2Item(armor.BlizzardArmlet),
            KH2Item(armor.BlizzaraArmlet),
            KH2Item(armor.BlizzagaArmlet),
            KH2Item(armor.BlizzagunArmlet),
            KH2Item(armor.ThunderTrinket),
            KH2Item(armor.ThundaraTrinket),
            KH2Item(armor.ThundagaTrinket),
            KH2Item(armor.ThundagunTrinket),
            KH2Item(armor.ShockCharm),
            KH2Item(armor.ShockCharmPlus, itemRarity.UNCOMMON),
            KH2Item(armor.ShadowAnklet),
            KH2Item(armor.DarkAnklet),
            KH2Item(armor.MidnightAnklet),
            KH2Item(armor.ChaosAnklet),
            KH2Item(armor.ChampionBelt),
            KH2Item(armor.AbasChain),
            KH2Item(armor.AegisChain),
            KH2Item(armor.Acrisius),
            KH2Item(armor.AcrisiusPlus),
            KH2Item(armor.CosmicChain),
            KH2Item(armor.PetiteRibbon, itemRarity.RARE),
            KH2Item(armor.Ribbon, itemRarity.RARE),
            KH2Item(armor.GrandRibbon, itemRarity.RARE),

            KH2Item(misc.MunnyPouchMickey, itemRarity.RARE),
            KH2Item(misc.MunnyPouchOlette, itemRarity.RARE),

            KH2Item(misc.HadesCupTrophy, itemRarity.UNCOMMON),
            KH2Item(misc.UnknownDisk, itemRarity.UNCOMMON),

            # KH2Item(misc.CrystalOrb), # removing for objectives
            KH2Item(misc.SeifersTrophy),
            KH2Item(misc.OlympusStone, itemRarity.UNCOMMON),
            KH2Item(misc.AuronsStatue),
            KH2Item(misc.CursedMedallion),
            KH2Item(misc.Present),
            KH2Item(misc.DecoyPresents),
            KH2Item(misc.StruggleTrophy),

            KH2Item(recipe.MegaRecipe),
            KH2Item(recipe.StarRecipe),
            KH2Item(recipe.RecoveryRecipe),
            KH2Item(recipe.SkillRecipe),
            KH2Item(recipe.GuardRecipe),
            KH2Item(recipe.RoadToDiscovery),
            KH2Item(recipe.StrengthBeyondStrength),
            KH2Item(recipe.BookOfShadows),
            KH2Item(recipe.CloakedThunder),
            KH2Item(recipe.EternalBlossom),
            KH2Item(recipe.RareDocument),
            KH2Item(recipe.StyleRecipe),
            KH2Item(recipe.MoonRecipe),
            KH2Item(recipe.QueenRecipe),
            KH2Item(recipe.KingRecipe),
            KH2Item(recipe.UltimateRecipe),

            KH2Item(map.TowerMap),
            KH2Item(map.TwilightTownMap),
            KH2Item(map.SunsetHillMap),
            KH2Item(map.MansionMap),
            KH2Item(map.CastlePerimeterMap),
            KH2Item(map.GreatMawMap),
            KH2Item(map.MarketplaceMap),
            KH2Item(map.DarkRemembranceMap),
            KH2Item(map.DepthsOfRemembranceMap),
            KH2Item(map.GardenOfAssemblageMap),
            KH2Item(map.CastleMap),
            KH2Item(map.BasementMap),
            KH2Item(map.CastleWallsMap),
            KH2Item(map.UnderworldMap),
            KH2Item(map.CavernsMap),
            KH2Item(map.ColiseumMap),
            KH2Item(map.CaveOfWondersMap),
            KH2Item(map.RuinsMap),
            KH2Item(map.AgrabahMap),
            KH2Item(map.PalaceMap),
            KH2Item(map.EncampmentAreaMap),
            KH2Item(map.VillageAreaMap),
            KH2Item(map.HundredAcreWoodMap),
            KH2Item(map.PigletsHowseMap),
            KH2Item(map.RabbitsHowseMap),
            KH2Item(map.KangasHowseMap),
            KH2Item(map.SpookyCaveMap),
            KH2Item(map.StarryHillMap),
            KH2Item(map.SavannahMap),
            KH2Item(map.PrideRockMap),
            KH2Item(map.OasisMap),
            KH2Item(map.UnderseaKingdomMap),
            KH2Item(map.DisneyCastleMap),
            KH2Item(map.CornerstoneHillMap),
            KH2Item(map.WindowOfTimeMap),
            KH2Item(map.LilliputMap),
            KH2Item(map.BuildingSiteMap),
            KH2Item(map.MickeysHouseMap),
            KH2Item(map.HalloweenTownMap),
            KH2Item(map.ChristmasTownMap),
            KH2Item(map.CurlyHillMap),
            KH2Item(map.NavalMap),
            KH2Item(map.IslaDeMuertaMap),
            KH2Item(map.ShipGraveyardMap),
            KH2Item(map.InterceptorMap),
            KH2Item(map.BlackPearlMap),
            KH2Item(map.PitCellAreaMap),
            KH2Item(map.IoTowerMap),
            KH2Item(map.CentralComputerCoreMap),
            KH2Item(map.SolarSailerSimulationMap),
            KH2Item(map.DarkCityMap),
            KH2Item(map.CastleThatNeverWasMap),
        ]

    @staticmethod
    def getStatItems() -> list[KH2Item]:
        """Get different dummy items that represent stat bonuses (Dummy 23, 24, 25, 26, 27, 16)"""
        return list(itertools.repeat(KH2Item(bonus.MaxHpUp), 20)) + \
            list(itertools.repeat(KH2Item(bonus.MaxMpUp), 4)) + \
            list(itertools.repeat(KH2Item(bonus.DriveGaugeUp, itemRarity.RARE), 6)) + \
            list(itertools.repeat(KH2Item(bonus.ArmorSlotUp, itemRarity.UNCOMMON), 3)) + \
            list(itertools.repeat(KH2Item(bonus.AccessorySlotUp, itemRarity.UNCOMMON), 3)) + \
            list(itertools.repeat(KH2Item(bonus.ItemSlotUp, itemRarity.UNCOMMON), 5))

    @staticmethod
    def getPromiseCharm() -> KH2Item:
        return KH2Item(misc.PromiseCharm, itemRarity.MYTHIC)

    @staticmethod
    def getDummyFormItems() -> list[KH2Item]:
        return [
            KH2Item(form.DummyFinalForm, Rarity=itemRarity.MYTHIC),
            KH2Item(form.DummyValorForm, Rarity=itemRarity.MYTHIC)
        ]
    
    @staticmethod
    def getFormToDummyMap() -> dict[int,int]:
        return {form.ValorForm.id:form.DummyValorForm.id,
                form.FinalForm.id:form.DummyFinalForm.id,}

    @staticmethod
    def getTT1Jailbreak() -> KH2Item:
        return KH2Item(misc.Poster)

    @staticmethod
    def getSupportAbilityList() -> list[KH2Item]:
        return [
            KH2Item(ability.Scan),
            KH2Item(ability.AerialRecovery),
            KH2Item(ability.ComboMaster, itemRarity.RARE),
            KH2Item(ability.ComboPlus),
            KH2Item(ability.ComboPlus),
            KH2Item(ability.AirComboPlus),
            KH2Item(ability.AirComboPlus),
            KH2Item(ability.ReactionBoost),
            KH2Item(ability.FinishingPlus, itemRarity.RARE),
            KH2Item(ability.FormBoost, itemRarity.UNCOMMON),
            KH2Item(ability.FormBoost, itemRarity.UNCOMMON),
            KH2Item(ability.SummonBoost),
            KH2Item(ability.Draw, itemRarity.UNCOMMON),
            KH2Item(ability.Draw, itemRarity.UNCOMMON),
            KH2Item(ability.LuckyLucky),
            KH2Item(ability.LuckyLucky),
            KH2Item(ability.LuckyLucky),
            KH2Item(ability.MpRage),
            KH2Item(ability.MpHaste),
            KH2Item(ability.MpHastera),
            KH2Item(ability.MpHastega),
            KH2Item(ability.Defender),
            KH2Item(ability.NoExperience),
        ]
    @staticmethod
    def getKeybladeAbilityList() -> list[KH2Item]:
        return [
            KH2Item(ability.Scan),
            KH2Item(ability.ComboPlus),
            KH2Item(ability.AirComboPlus),
            KH2Item(ability.ComboBoost, itemRarity.UNCOMMON),
            KH2Item(ability.AirComboBoost, itemRarity.UNCOMMON),
            KH2Item(ability.ReactionBoost),
            KH2Item(ability.FinishingPlus, itemRarity.RARE),
            KH2Item(ability.NegativeCombo, itemRarity.RARE),
            KH2Item(ability.BerserkCharge, itemRarity.UNCOMMON),
            KH2Item(ability.DamageDrive),
            KH2Item(ability.DriveBoost),
            KH2Item(ability.FormBoost, itemRarity.UNCOMMON),
            KH2Item(ability.ExperienceBoost, itemRarity.RARE),
            KH2Item(ability.Draw, itemRarity.UNCOMMON),
            KH2Item(ability.Jackpot),
            KH2Item(ability.DriveConverter, itemRarity.UNCOMMON),
            KH2Item(ability.FireBoost),
            KH2Item(ability.BlizzardBoost),
            KH2Item(ability.ThunderBoost),
            KH2Item(ability.ItemBoost, itemRarity.UNCOMMON),
            KH2Item(ability.MpRage),
            KH2Item(ability.MpHaste),
            KH2Item(ability.MpHastera),
            KH2Item(ability.MpHastega),
            KH2Item(ability.DamageControl),
            KH2Item(ability.NoExperience),
            KH2Item(ability.LightAndDarkness, itemRarity.RARE),
        ]

    @staticmethod
    def getLevelAbilityList() -> list[KH2Item]:
        return [
            KH2Item(ability.ComboBoost, itemRarity.UNCOMMON),
            KH2Item(ability.ExperienceBoost, itemRarity.RARE),
            KH2Item(ability.MagicLockOn),
            KH2Item(ability.ReactionBoost),
            KH2Item(ability.ItemBoost, itemRarity.UNCOMMON),
            KH2Item(ability.LeafBracer),
            KH2Item(ability.FireBoost),
            KH2Item(ability.DriveBoost),
            KH2Item(ability.Draw, itemRarity.UNCOMMON),
            KH2Item(ability.CombinationBoost),
            KH2Item(ability.DamageDrive),
            KH2Item(ability.AirComboBoost, itemRarity.UNCOMMON),
            KH2Item(ability.BlizzardBoost),
            KH2Item(ability.DriveConverter, itemRarity.UNCOMMON),
            KH2Item(ability.NegativeCombo, itemRarity.RARE),
            KH2Item(ability.OnceMore, itemRarity.MYTHIC),
            KH2Item(ability.FinishingPlus, itemRarity.RARE),
            KH2Item(ability.ThunderBoost),
            KH2Item(ability.Defender),
            KH2Item(ability.BerserkCharge, itemRarity.UNCOMMON),
            KH2Item(ability.Jackpot),
            KH2Item(ability.SecondChance, itemRarity.MYTHIC),
            KH2Item(ability.DamageControl),
        ]

    @staticmethod
    def getActionAbilityList() -> list[KH2Item]:
        return [
            KH2Item(ability.Guard),
            KH2Item(ability.UpperSlash),
            KH2Item(ability.HorizontalSlash, itemRarity.RARE),
            KH2Item(ability.FinishingLeap, itemRarity.RARE),
            KH2Item(ability.RetaliatingSlash),
            KH2Item(ability.Slapshot, itemRarity.UNCOMMON),
            KH2Item(ability.DodgeSlash),
            KH2Item(ability.FlashStep, itemRarity.RARE),
            KH2Item(ability.SlideDash, itemRarity.RARE),
            KH2Item(ability.VicinityBreak),
            KH2Item(ability.GuardBreak, itemRarity.RARE),
            KH2Item(ability.Explosion, itemRarity.RARE),
            KH2Item(ability.AerialSweep, itemRarity.UNCOMMON),
            KH2Item(ability.AerialDive, itemRarity.RARE),
            KH2Item(ability.AerialSpiral, itemRarity.RARE),
            KH2Item(ability.AerialFinish),
            KH2Item(ability.MagnetBurst, itemRarity.RARE),
            KH2Item(ability.Counterguard),
            KH2Item(ability.AutoValor, itemRarity.UNCOMMON),
            KH2Item(ability.AutoWisdom, itemRarity.UNCOMMON),
            KH2Item(ability.AutoLimitForm, itemRarity.UNCOMMON),
            KH2Item(ability.AutoMaster, itemRarity.UNCOMMON),
            KH2Item(ability.AutoFinal, itemRarity.UNCOMMON),
            KH2Item(ability.AutoSummon, itemRarity.UNCOMMON),
            KH2Item(ability.TrinityLimit, itemRarity.RARE),
        ]

    @staticmethod
    def sort_ability_items(abilities: Iterable[KH2Item]) -> list[KH2Item]:
        def key_function(ability_item: KH2Item) -> int:
            inventory_item = ability_item.item
            if isinstance(inventory_item, Ability):
                return inventory_item.sort_index
            else:
                return 999999
        return sorted(abilities, key=key_function)

    @staticmethod
    def getNullItem() -> KH2Item:
        return KH2Item(misc.NullItem)

    # @staticmethod
    # def sharedMultiItem() -> KH2Item:
    #     return KH2Item(misc.SharedMultiworldItem)
    @staticmethod
    def objectiveReportItem() -> KH2Item:
        return KH2Item(misc.ObjectiveReport)


    @staticmethod
    def objectiveItem() -> KH2Item:
        return KH2Item(misc.ObjectiveItem)
    
    @staticmethod
    def emblemItem() -> KH2Item:
        return KH2Item(misc.EmblemItem)

    @staticmethod
    def weaponslot_id_to_keyblade_item(location_id: int) -> Optional[Keyblade]:
        slot_id_to_keyblade = {
            weaponslot.LocationId.KingdomKeyD: None,
            weaponslot.LocationId.AlphaWeapon: None,
            weaponslot.LocationId.OmegaWeapon: None,
            weaponslot.LocationId.KingdomKey: None,
            weaponslot.LocationId.Oathkeeper: keyblade.Oathkeeper,
            weaponslot.LocationId.Oblivion: keyblade.Oblivion,
            weaponslot.LocationId.StarSeeker: keyblade.StarSeeker,
            weaponslot.LocationId.HiddenDragon: keyblade.HiddenDragon,
            weaponslot.LocationId.HerosCrest: keyblade.HerosCrest,
            weaponslot.LocationId.Monochrome: keyblade.Monochrome,
            weaponslot.LocationId.FollowTheWind: keyblade.FollowTheWind,
            weaponslot.LocationId.CircleOfLife: keyblade.CircleOfLife,
            weaponslot.LocationId.PhotonDebugger: keyblade.PhotonDebugger,
            weaponslot.LocationId.GullWing: keyblade.GullWing,
            weaponslot.LocationId.RumblingRose: keyblade.RumblingRose,
            weaponslot.LocationId.GuardianSoul: keyblade.GuardianSoul,
            weaponslot.LocationId.WishingLamp: keyblade.WishingLamp,
            weaponslot.LocationId.DecisivePumpkin: keyblade.DecisivePumpkin,
            weaponslot.LocationId.SweetMemories: keyblade.SweetMemories,
            weaponslot.LocationId.MysteriousAbyss: keyblade.MysteriousAbyss,
            weaponslot.LocationId.SleepingLion: keyblade.SleepingLion,
            weaponslot.LocationId.BondOfFlame: keyblade.BondOfFlame,
            weaponslot.LocationId.TwoBecomeOne: keyblade.TwoBecomeOne,
            weaponslot.LocationId.FatalCrest: keyblade.FatalCrest,
            weaponslot.LocationId.Fenrir: keyblade.Fenrir,
            weaponslot.LocationId.UltimaWeapon: keyblade.UltimaWeapon,
            weaponslot.LocationId.WinnersProof: keyblade.WinnersProof,
            weaponslot.LocationId.Pureblood: keyblade.Pureblood
        }
        return slot_id_to_keyblade[location_id]

    @staticmethod
    def getJunkList(betterJunk) -> list[KH2Item]:
        if betterJunk:
            return [
                KH2Item(consumable.Potion),
                KH2Item(consumable.HiPotion),
                KH2Item(consumable.Ether),
                KH2Item(consumable.Elixir, itemRarity.UNCOMMON),
                KH2Item(consumable.MegaPotion),
                KH2Item(consumable.MegaEther),
                KH2Item(consumable.Megalixir, itemRarity.UNCOMMON),
                KH2Item(consumable.Tent),
                KH2Item(consumable.DriveRecovery, itemRarity.UNCOMMON),
                KH2Item(consumable.HighDriveRecovery, itemRarity.UNCOMMON),
                KH2Item(consumable.PowerBoost),
                KH2Item(consumable.MagicBoost),
                KH2Item(consumable.DefenseBoost),
                KH2Item(consumable.ApBoost),
            ]
        return [
            KH2Item(synth.BlazingShard),
            KH2Item(synth.BlazingStone),
            KH2Item(synth.BlazingGem),
            KH2Item(synth.BlazingCrystal),
            KH2Item(synth.FrostShard),
            KH2Item(synth.FrostStone),
            KH2Item(synth.FrostGem),
            KH2Item(synth.FrostCrystal),
            KH2Item(synth.LightningShard),
            KH2Item(synth.LightningStone),
            KH2Item(synth.LightningGem),
            KH2Item(synth.LightningCrystal),
            KH2Item(synth.LucidShard),
            KH2Item(synth.LucidStone),
            KH2Item(synth.LucidGem),
            KH2Item(synth.LucidCrystal),
            KH2Item(synth.PowerShard),
            KH2Item(synth.PowerStone),
            KH2Item(synth.PowerGem),
            KH2Item(synth.PowerCrystal),
            KH2Item(synth.DarkShard),
            KH2Item(synth.DarkStone),
            KH2Item(synth.DarkGem),
            KH2Item(synth.DarkCrystal),
            KH2Item(synth.DenseShard),
            KH2Item(synth.DenseStone),
            KH2Item(synth.DenseGem),
            KH2Item(synth.DenseCrystal),
            KH2Item(synth.TwilightShard),
            KH2Item(synth.TwilightStone),
            KH2Item(synth.TwilightGem),
            KH2Item(synth.TwilightCrystal),
            KH2Item(synth.MythrilShard),
            KH2Item(synth.MythrilStone),
            KH2Item(synth.MythrilGem),
            KH2Item(synth.MythrilCrystal),
            KH2Item(synth.RemembranceShard),
            KH2Item(synth.RemembranceStone),
            KH2Item(synth.RemembranceGem),
            KH2Item(synth.RemembranceCrystal),
            KH2Item(synth.TranquilityShard),
            KH2Item(synth.TranquilityStone),
            KH2Item(synth.TranquilityGem),
            KH2Item(synth.TranquilityCrystal),
            KH2Item(synth.BrightShard),
            KH2Item(synth.BrightStone),
            KH2Item(synth.BrightGem),
            KH2Item(synth.BrightCrystal),
            KH2Item(synth.EnergyShard),
            KH2Item(synth.EnergyStone),
            KH2Item(synth.EnergyGem),
            KH2Item(synth.EnergyCrystal),
            KH2Item(synth.SerenityShard),
            KH2Item(synth.SerenityStone),
            KH2Item(synth.SerenityGem),
            KH2Item(synth.SerenityCrystal),
            KH2Item(synth.LostIllusion),
            KH2Item(synth.ManifestIllusion),
            KH2Item(synth.Orichalcum),
            KH2Item(synth.OrichalcumPlus),
            KH2Item(consumable.Potion),
            KH2Item(consumable.HiPotion),
            KH2Item(consumable.Ether),
            KH2Item(consumable.Elixir, itemRarity.UNCOMMON),
            KH2Item(consumable.MegaPotion),
            KH2Item(consumable.MegaEther),
            KH2Item(consumable.Megalixir, itemRarity.UNCOMMON),
            KH2Item(consumable.Tent),
            KH2Item(consumable.DriveRecovery, itemRarity.UNCOMMON),
            KH2Item(consumable.HighDriveRecovery, itemRarity.UNCOMMON),
            KH2Item(consumable.PowerBoost),
            KH2Item(consumable.MagicBoost),
            KH2Item(consumable.DefenseBoost),
            KH2Item(consumable.ApBoost),
        ]

    @staticmethod
    def getSynthRequirementsList() -> list[KH2Item]:
        return [
            KH2Item(synth.BlazingShard),
            KH2Item(synth.BlazingStone),
            KH2Item(synth.BlazingGem),
            KH2Item(synth.BlazingCrystal),
            KH2Item(synth.FrostShard),
            KH2Item(synth.FrostStone),
            KH2Item(synth.FrostGem),
            KH2Item(synth.FrostCrystal),
            KH2Item(synth.LightningShard),
            KH2Item(synth.LightningStone),
            KH2Item(synth.LightningGem),
            KH2Item(synth.LightningCrystal),
            KH2Item(synth.LucidShard),
            KH2Item(synth.LucidStone),
            KH2Item(synth.LucidGem),
            KH2Item(synth.LucidCrystal),
            KH2Item(synth.PowerShard),
            KH2Item(synth.PowerStone),
            KH2Item(synth.PowerGem),
            KH2Item(synth.PowerCrystal),
            KH2Item(synth.DarkShard),
            KH2Item(synth.DarkStone),
            KH2Item(synth.DarkGem),
            KH2Item(synth.DarkCrystal),
            KH2Item(synth.DenseShard),
            KH2Item(synth.DenseStone),
            KH2Item(synth.DenseGem),
            KH2Item(synth.DenseCrystal),
            KH2Item(synth.TwilightShard),
            KH2Item(synth.TwilightStone),
            KH2Item(synth.TwilightGem),
            KH2Item(synth.TwilightCrystal),
            # may be able to bring these back if the class of item is known
            # KH2Item(synth.BrightShard),
            # KH2Item(synth.BrightStone),
            # KH2Item(synth.BrightGem),
            # KH2Item(synth.BrightCrystal),
            # KH2Item(synth.EnergyShard),
            # KH2Item(synth.EnergyStone),
            # KH2Item(synth.EnergyGem),
            # KH2Item(synth.EnergyCrystal),
        ]

    @staticmethod
    def donald_ability_list() -> list[KH2Item]:
        return [
            KH2Item(DonaldAbilities.DonaldFire),
            KH2Item(DonaldAbilities.DonaldBlizzard),
            KH2Item(DonaldAbilities.DonaldThunder),
            KH2Item(DonaldAbilities.DonaldCure),
            KH2Item(DonaldAbilities.Fantasia),
            KH2Item(DonaldAbilities.FlareForce),
            KH2Item(DonaldAbilities.MpRage),
            KH2Item(DonaldAbilities.Jackpot),
            KH2Item(DonaldAbilities.LuckyLucky),
            KH2Item(DonaldAbilities.FireBoost),
            KH2Item(DonaldAbilities.BlizzardBoost),
            KH2Item(DonaldAbilities.ThunderBoost),
            KH2Item(DonaldAbilities.FireBoost),
            KH2Item(DonaldAbilities.BlizzardBoost),
            KH2Item(DonaldAbilities.ThunderBoost),
            KH2Item(DonaldAbilities.MpRage),
            KH2Item(DonaldAbilities.MpHastera),
            KH2Item(DonaldAbilities.AutoLimitParty),
            KH2Item(DonaldAbilities.HyperHealing),
            KH2Item(DonaldAbilities.AutoHealing),
            KH2Item(DonaldAbilities.MpHastega),
            KH2Item(DonaldAbilities.ItemBoost),
            KH2Item(DonaldAbilities.DamageControl),
            KH2Item(DonaldAbilities.HyperHealing),
            KH2Item(DonaldAbilities.MpRage),
            KH2Item(DonaldAbilities.MpHaste),
            KH2Item(DonaldAbilities.MpHastera),
            KH2Item(DonaldAbilities.MpHastega),
            KH2Item(DonaldAbilities.MpHaste),
            KH2Item(DonaldAbilities.DamageControl),
            KH2Item(DonaldAbilities.MpHastera),
            KH2Item(DonaldAbilities.Draw),
        ]

    @staticmethod
    def goofy_ability_list() -> list[KH2Item]:
        return [
            KH2Item(GoofyAbilities.GoofyTornado),
            KH2Item(GoofyAbilities.GoofyTurbo),
            KH2Item(GoofyAbilities.GoofyBash),
            KH2Item(GoofyAbilities.TornadoFusion),
            KH2Item(GoofyAbilities.Teamwork),
            KH2Item(GoofyAbilities.Draw),
            KH2Item(GoofyAbilities.Jackpot),
            KH2Item(GoofyAbilities.LuckyLucky),
            KH2Item(GoofyAbilities.ItemBoost),
            KH2Item(GoofyAbilities.MpRage),
            KH2Item(GoofyAbilities.Defender),
            KH2Item(GoofyAbilities.DamageControl),
            KH2Item(GoofyAbilities.AutoLimitParty),
            KH2Item(GoofyAbilities.SecondChance),
            KH2Item(GoofyAbilities.OnceMore),
            KH2Item(GoofyAbilities.AutoChange),
            KH2Item(GoofyAbilities.HyperHealing),
            KH2Item(GoofyAbilities.AutoHealing),
            KH2Item(GoofyAbilities.Defender),
            KH2Item(GoofyAbilities.HyperHealing),
            KH2Item(GoofyAbilities.MpHaste),
            KH2Item(GoofyAbilities.MpHastera),
            KH2Item(GoofyAbilities.MpRage),
            KH2Item(GoofyAbilities.MpHastega),
            KH2Item(GoofyAbilities.ItemBoost),
            KH2Item(GoofyAbilities.DamageControl),
            KH2Item(GoofyAbilities.Protect),
            KH2Item(GoofyAbilities.Protera),
            KH2Item(GoofyAbilities.Protega),
            KH2Item(GoofyAbilities.DamageControl),
            KH2Item(GoofyAbilities.Protect),
            KH2Item(GoofyAbilities.Protera),
            KH2Item(GoofyAbilities.Protega),
        ]



from dataclasses import dataclass, field

from List.configDict import itemType, locationType, locationCategory


@dataclass(unsafe_hash=True)
class KH2Location:
    """ Location description class for the properties of a location """
    LocationId: int
    Description: str
    LocationCategory: locationCategory
    LocationTypes: list[locationType] = field(compare=False)
    InvalidChecks: list[itemType] = field(default_factory=list, compare=False)
    VanillaItems: list[int] = field(default_factory=list, compare=False)

    def name(self) -> str:
        return self.Description

    def __eq__(self, obj):
        return self.LocationId == obj.LocationId and self.LocationCategory == obj.LocationCategory and self.LocationTypes == obj.LocationTypes

    def __str__(self):
        return f"{self.LocationTypes} - {self.Description}"



import io
import textwrap
import zipfile
from dataclasses import dataclass
from typing import Optional, Any
from zipfile import ZipFile, ZIP_DEFLATED

import yaml

from Class.openkhmod import (
    AttackEntriesOrganizer,
    ModYml,
    Bonuses,
    FormLevels,
    Items,
    LevelUps,
    Messages,
    PlayerParams,
    PrizeTable,
    Treasures,
    write_yaml_to_zip_file,
    Asset,
)
from Module.RandomizerSettings import RandomizerSettings
from Module.resources import resource_path


def _relative_mod_file(name: str) -> str:
    """Used to help organize mod-specific files inside the mod zip."""
    return f"randoseed-mod-files/{name}"


@dataclass(frozen=True)
class ChestVisualAssignment:
    chest_index: int
    spawn_file_path: str
    chest_visual_id: int


class SeedModBuilder:
    def __init__(self, title: str, out_zip: ZipFile):
        self.out_zip = out_zip
        self.mod_yml = ModYml(
            title, description="Generated by the KH2 Randomizer Seed Generator."
        )
        self.bonuses = Bonuses(_relative_mod_file("BonsList.yml"))
        self.form_levels = FormLevels(_relative_mod_file("FmlvList.yml"))
        self.items = Items(_relative_mod_file("ItemList.yml"))
        self.level_ups = LevelUps(_relative_mod_file("LvupList.yml"))
        self.messages = Messages(_relative_mod_file("sys.yml"))
        self.journal_txt = Messages(_relative_mod_file("jm.yml"), unicode_output=True)
        self.player_params = PlayerParams(_relative_mod_file("PlrpList.yml"))
        self.prize_table = PrizeTable(_relative_mod_file("przt.yml"))
        self.treasures = Treasures(_relative_mod_file("TrsrList.yml"))
        self.atkp_organizer = AttackEntriesOrganizer(_relative_mod_file("AtkpList.yml"))

    def add_base_assets(self):
        """Adds asset entries to the mod for files that get included with every seed."""
        self.mod_yml.add_assets(
            [
                {
                    "name": "msg/us/sys.bar",
                    "multi": [
                        {"name": "msg/fr/sys.bar"},
                        {"name": "msg/gr/sys.bar"},
                        {"name": "msg/it/sys.bar"},
                        {"name": "msg/sp/sys.bar"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "sys",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.messages.source_name, "language": "en"}
                            ],
                        }
                    ],
                },
                {
                    "name": "msg/jp/sys.bar",
                    "platform": "ps2",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "sys",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.messages.source_name, "language": "en"}
                            ],
                        }
                    ],
                },
                {
                    "name": "msg/jp/sys.bar",
                    "platform": "pc",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "sys",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.messages.source_name, "language": "jp"}
                            ],
                        }
                    ],
                },
                {
                    "name": "msg/us/jm.bar",
                    "multi": [
                        {"name": "msg/fr/jm.bar"},
                        {"name": "msg/gr/jm.bar"},
                        {"name": "msg/it/jm.bar"},
                        {"name": "msg/sp/jm.bar"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "jm",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.journal_txt.source_name, "language": "en"}
                            ],
                        }
                    ],
                },
                {
                    "name": "msg/jp/jm.bar",
                    "platform": "ps2",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "jm",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.journal_txt.source_name, "language": "en"}
                            ],
                        }
                    ],
                },
                {
                    "name": "msg/jp/jm.bar",
                    "platform": "pc",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "jm",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {"name": self.journal_txt.source_name, "language": "jp"}
                            ],
                        }
                    ],
                },
                {
                    "name": "00battle.bin",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "fmlv",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.form_levels.source_name, "type": "fmlv"}
                            ],
                        },
                        {
                            "name": "lvup",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.level_ups.source_name, "type": "lvup"}
                            ],
                        },
                        {
                            "name": "bons",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.bonuses.source_name, "type": "bons"}
                            ],
                        },
                        {
                            "name": "plrp",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.player_params.source_name, "type": "plrp"}
                            ],
                        },
                    ],
                },
                {
                    "name": "03system.bin",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "trsr",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.treasures.source_name, "type": "trsr"}
                            ],
                        },
                        {
                            "name": "item",
                            "method": "listpatch",
                            "type": "List",
                            "source": [
                                {"name": self.items.source_name, "type": "item"}
                            ],
                        },
                    ],
                },
            ]
        )
    def add_emblem_randomization_mods(self, num_emblems_needed):
        self.messages.add_message(
            15115,
            en="Lucky Emblem",
            jp="Lucky Emblem", # TODO make actual text for this
        )
        self.messages.add_message(
            15116,
            en=f"A Lucky Emblem.\nThe King left these for you to find.\n{num_emblems_needed} of these are required to open the door...",
            jp=f"A Lucky Emblem.\nThe King left these for you to find.\n{num_emblems_needed} of these are required to open the door...", # TODO make actual text for this
        )
        mark_image_source = _relative_mod_file("objectives/completionmark.dds")
        emu_mark_image_source = _relative_mod_file("objectives/completionmark.imd")
        self.mod_yml.add_asset({
            "name": "remastered/itempic/item-226.imd/-0.dds",
            "method": "copy",
            "platform": "pc",
            "source": [
                {
                    "name": mark_image_source,
                }
            ]
        })
        self.mod_yml.add_asset({
            "name": "itempic/item-226.imd",
            "method": "copy",
            "platform": "ps2",
            "source": [
                {
                    "name": emu_mark_image_source,
                }
            ]
        })
        self.out_zip.write(
            resource_path("static/objectives/completionmark.dds"), mark_image_source
        )
        self.out_zip.write(
            resource_path("static/objectives/completionmark.imd"), emu_mark_image_source
        )


    def add_objective_randomization_mods(self, num_objectives_needed, objective_list):
        def convert_string_to_unicode(string: str, newlines: int = 0):
            return (
                    "".join(r"\u{:04X}".format(ord(character)) for character in textwrap.fill(string, width=30))
                    + "NEWLINE" * newlines
            )
        self.messages.add_message(
            15115,
            en="Completion Mark",
            jp="Completion Mark", # TODO make actual text for this
        )
        self.messages.add_message(
            15116,
            en=f"An Objective Completion Mark.\nAwarded to those who complete given tasks.\n{num_objectives_needed} of these are required to open the door...",
            jp=f"An Objective Completion Mark.\nAwarded to those who complete given tasks.\n{num_objectives_needed} of these are required to open the door...", # TODO make actual text for this
        )
        self.journal_txt.add_message(
            14260,
            en=convert_string_to_unicode("Objective List")
        )       
        self.journal_txt.add_message(
            13945,
            en=convert_string_to_unicode("Reports/Objectives")
        )     
        listText = ''
        counter = 1
        for task in objective_list:
            listText += convert_string_to_unicode(str(counter)+'. '+task.Name,1)
            counter+=1

        self.journal_txt.add_message(
            message_id=14261,
            en=listText
        )
        
        mark_image_source = _relative_mod_file("objectives/completionmark.dds")
        emu_mark_image_source = _relative_mod_file("objectives/completionmark.imd")
        report_addition_bin = _relative_mod_file("objectives/ansem_modified.bin")

        self.mod_yml.add_asset({
            "name": "remastered/itempic/item-226.imd/-0.dds",
            "method": "copy",
            "platform": "pc",
            "source": [
                {
                    "name": mark_image_source,
                }
            ]
        })
        self.mod_yml.add_asset({
            "name": "itempic/item-226.imd",
            "method": "copy",
            "platform": "ps2",
            "source": [
                {
                    "name": emu_mark_image_source,
                }
            ]
        })

        source_data = {
                    "name": "anse",
                    "method":"copy",
                    "type":"jimidata",
                    "source" : [{"name":report_addition_bin}]
                }

        if self.mod_yml.find_asset("menu/fm/jiminy.bar") is not None:
            self.mod_yml.add_asset_source("menu/fm/jiminy.bar",source_data)
            self.mod_yml.add_asset_source("menu/us/jiminy.bar",source_data)
        else:
            self.mod_yml.add_asset(
                {"name": "menu/fm/jiminy.bar", "method": "binarc", "source": [source_data]}
            )
            self.mod_yml.add_asset(
                {
                    "name": "menu/us/jiminy.bar",
                    "platform": "pc",
                    "multi": [
                        {"name": "menu/fr/jiminy.bar"},
                        {"name": "menu/gr/jiminy.bar"},
                        {"name": "menu/it/jiminy.bar"},
                        {"name": "menu/sp/jiminy.bar"},
                        {"name": "menu/uk/jiminy.bar"},
                    ],
                    "method": "binarc",
                    "source": [source_data],
                }
            )
        self.out_zip.write(
            resource_path("static/objectives/completionmark.dds"), mark_image_source
        )
        self.out_zip.write(
            resource_path("static/objectives/completionmark.imd"), emu_mark_image_source
        )
        self.out_zip.write(
            resource_path("static/objectives/ansem_modified.bin"), report_addition_bin
        )


    def add_sora_bdscript(self,item_ids,ability_ids,equipped):
        if len(ability_ids)==0 and len(item_ids)==0:
            return
        preamble = "pushImm 1\n" \
        "syscall 0, 61 ; trap_saveram_get_partram (1 in, 1 out)\n" \
        "popToSp 0\n"
        ability_string = "pushFromFSp 0\n"\
        "pushImm ITEM_ID\n"\
        "pushImm EQUIPPED\n"\
        "syscall 0, 99 ; trap_partram_add_ability (3 in, 0 out)\n"
        item_string = "pushImm ITEM_ID\n"\
        "pushImm 100\n"\
        "syscall 0, 63 ; trap_item_get (2 in, 0 out)\n"
        postamble = "ret\n"

        item_script = ""
        if len(ability_ids)!=0:
            item_script+=preamble
            for id in ability_ids:
                item_script+=ability_string.replace("ITEM_ID",str(id)).replace("EQUIPPED",str(int(equipped)))
        if len(item_ids)!=0:
            for id in item_ids:
                item_script+=item_string.replace("ITEM_ID",str(id))
        item_script+=postamble

        with open(resource_path("static/starting/starting_inventory.bdscript"),"r") as infile:
            full_item_script = infile.read()

        full_item_script = full_item_script.replace("ITEM_LIST_GOES_HERE",item_script)
        starting_item_bdscript = _relative_mod_file("starting_items/starting_inventory.bdscript")
        starting_item_evt = _relative_mod_file("starting_items/starting_item_event.script")

        self.mod_yml.add_asset({
            "name": "ard/tt32.ard",
            "method": "binarc",
            "multi": [
                {"name":"ard/us/tt32.ard"},
                {"name":"ard/fr/tt32.ard"},
                {"name":"ard/gr/tt32.ard"},
                {"name":"ard/it/tt32.ard"},
                {"name":"ard/sp/tt32.ard"},
                {"name":"ard/uk/tt32.ard"},
                {"name":"ard/jp/tt32.ard"},
            ],
            "source": [
                {
                    "name": "evt",
                    "method":"areadatascript",
                    "type":"areadatascript",
                    "source" : [{"name":starting_item_evt}]

                },
                {
                    "name": "starting_abilities",
                    "method":"bdscript",
                    "type":"Bdx",
                    "source" : [{"name":starting_item_bdscript}]

                }
            ],
        })
        
        self.out_zip.writestr(starting_item_bdscript, full_item_script)
        self.out_zip.write(
            resource_path("static/starting/starting_item_event.script"), starting_item_evt
        )

    def add_base_messages(
            self,
            seed_hash_icons: list[str],
            crit_mode: bool,
            final_door_requirement_message: str
    ):
        """Adds messages that are included with every seed."""
        seed_hash_string = " ".join(
            ["{:icon " + icon + "}" for icon in seed_hash_icons]
        )
        self.messages.add_message(17198, en=seed_hash_string, jp=seed_hash_string)

        self.messages.add_message(20239, en=final_door_requirement_message, jp=final_door_requirement_message)

        # Not needed until we figure out how to display the correct amount
        # self.messages.add_message(19482, en="Important Checks Found")

        if crit_mode:
            self.messages.add_message(
                17201,
                en="{:color #FF000080}Beginner (WARNING)",
                jp="{:color #FF000080}ビギナーモード (注意!)",
            )
            self.messages.add_message(
                17202,
                en="{:color #FF000080}Standard (WARNING)",
                jp="{:color #FF000080}スタンダードモード (注意!)",
            )
            self.messages.add_message(
                17203,
                en="{:color #FF000080}Proud (WARNING)",
                jp="{:color #FF000080}プラウドモード (注意!)",
            )
            self.messages.add_message(
                17204,
                en="An easier mode for beginners.\n{:color #FF000080}Critical Bonuses are turned on. The seven\nrandom starting items will be unobtainable.",
                jp="敵が弱く サクサクすすめるかんたんモードです\n{:color #FF000080}クリティカル特典が有効になっていますが、このモード\nではゲーム開始時に7つの特典を受け取れません",
            )
            self.messages.add_message(
                17205,
                en="A balanced mode best for those challenging\nthis game for the first time.\n{:color #FF000080}Critical Bonuses are turned on. The seven\nrandom starting items will be unobtainable.",
                jp="初プレイにピッタリの ほどよいバランスのモードです\n{:color #FF000080}クリティカル特典が有効になっていますが、このモード\nではゲーム開始時に7つの特典を受け取れません",
            )
            self.messages.add_message(
                17206,
                en="A difficult mode with stronger enemies.\nBest for those seeking a challenge.\n{:color #FF000080}Critical Bonuses are turned on. The seven\nrandom starting items will be unobtainable.",
                jp="敵が強く スリリングなバトルがたのしめるモードです\n{:color #FF000080}クリティカル特典が有効になっていますが、このモード\nではゲーム開始時に7つの特典を受け取れません",
            )
            self.messages.add_message(
                20020,
                en="A true test of skill for the adept. Begin\nwith certain abilities and other perks.\n{:color #F0F00080}Critical Bonuses are turned on. The seven\nstarting items have been randomized.",
                jp="アクションのウデがためされる 上級者向けのモードです\n{:color #FF000080}クリティカル特典が有効になっています。ゲーム開始時\nにランダムな7つの特典を受け取れます",
            )
        else:
            self.messages.add_message(
                20020,
                en="A true test of skill for the adept. Begin\nwith certain abilities and other perks.\n{:color #F0F00080}Critical Bonuses are turned off. The seven\nstarting items will be junk.",
                jp="アクションのウデがためされる 上級者向けのモードです\n{:color #FF000080}クリティカル特典が無効になっています。ゲーム開始時\nにランダムな7つのジャンクアイテムを受け取れます",
            )

    def write_as_data_split_assets(self):
        """Adds assets and files to the mod for AS/Data split."""
        for as_room in ["hb32", "hb33", "hb34", "hb38"]:
            script_name = f"{as_room}evt.script"
            source_name = _relative_mod_file(f"as_data_split/{script_name}")
            self._add_areadata_script_asset(ard_id=as_room, source_name=source_name)
            self.out_zip.write(
                resource_path(f"static/as_data_split/{script_name}"), source_name
            )

    def write_puzzle_assets(self, modified_puzzle_binary: bytearray):
        """Adds assets and files to the mod for modified puzzle rewards."""
        source_name = _relative_mod_file("modified_puzzle.bin")

        source_list = [
            {
                "name": "puzz",
                "type": "jimidata",
                "method": "copy",
                "source": [{"name": source_name}],
            }
        ]
        self.mod_yml.add_asset(
            {"name": "menu/fm/jiminy.bar", "method": "binarc", "source": source_list}
        )
        self.mod_yml.add_asset(
            {
                "name": "menu/us/jiminy.bar",
                "platform": "pc",
                "multi": [
                    {"name": "menu/fr/jiminy.bar"},
                    {"name": "menu/gr/jiminy.bar"},
                    {"name": "menu/it/jiminy.bar"},
                    {"name": "menu/sp/jiminy.bar"},
                    {"name": "menu/uk/jiminy.bar"},
                ],
                "method": "binarc",
                "source": source_list,
            }
        )

        self.out_zip.writestr(source_name, modified_puzzle_binary)

    def write_shop_assets(self, modified_shop_binary: bytearray):
        """Adds assets and files to the mod for modified shop content."""
        source_name = _relative_mod_file("modified_shop.bin")

        self.mod_yml.add_asset_source(
            "03system.bin",
            {
                "name": "shop",
                "type": "unknown41",
                "method": "copy",
                "source": [{"name": source_name}],
            },
        )

        self.out_zip.writestr(source_name, modified_shop_binary)

    def write_synth_assets(
        self,
        modified_recipes_binary: bytearray,
        modified_requirements_binary: bytearray,
    ):
        """Adds assets and files to the mod for modified synthesis recipes and requirements."""
        recipes_source_name = _relative_mod_file("modified_synth.bin")
        requirements_source_name = _relative_mod_file("modified_synth_reqs.bin")

        source_list = [
            {
                "name": "reci",
                "type": "synthesis",
                "method": "copy",
                "source": [{"name": recipes_source_name}],
            },
            {
                "name": "cond",
                "type": "synthesis",
                "method": "copy",
                "source": [{"name": requirements_source_name}],
            },
        ]
        self.mod_yml.add_asset(
            {"name": "menu/fm/mixdata.bar", "method": "binarc", "source": source_list}
        )
        self.mod_yml.add_asset(
            {
                "name": "menu/us/mixdata.bar",
                "platform": "pc",
                "multi": [
                    {"name": "menu/fr/mixdata.bar"},
                    {"name": "menu/gr/mixdata.bar"},
                    {"name": "menu/it/mixdata.bar"},
                    {"name": "menu/sp/mixdata.bar"},
                    {"name": "menu/uk/mixdata.bar"},
                    {"name": "menu/jp/mixdata.bar"},
                ],
                "method": "binarc",
                "source": source_list,
            }
        )

        self.out_zip.writestr(recipes_source_name, modified_recipes_binary)
        self.out_zip.writestr(requirements_source_name, modified_requirements_binary)

    def write_skip_carpet_escape_assets(self):
        """Adds assets and files to the mod for magic carpet escape."""
        source_name = _relative_mod_file("skip_carpet_escape.script")
        self._add_areadata_script_asset(ard_id="al11", source_name=source_name)
        self.out_zip.write(
            resource_path("static/skip_carpet_escape.script"), source_name
        )

    def write_block_cor_skip_assets(self):
        """Adds assets and files to the mod for blocking CoR skip."""
        source_name = _relative_mod_file("disable_cor_skip.script")
        self._add_areadata_script_asset(ard_id="hb24", source_name=source_name)
        self.out_zip.write(resource_path("static/disable_cor_skip.script"), source_name)

    def write_block_shan_yu_skip_assets(self):
        """Adds assets and files to the mod for blocking Shan-Yu skip."""
        source_name = _relative_mod_file("disable_shan_yu_skip.script")
        self._add_areadata_script_asset(ard_id="mu09", source_name=source_name)
        self.out_zip.write(
            resource_path("static/disable_shan_yu_skip.script"), source_name
        )

    def write_atlantica_tutorial_skip_assets(self):
        """Adds assets and files to the mod for Atlantica tutorial skip."""
        source_name = _relative_mod_file("atlantica_skip.script")
        self._add_areadata_script_asset(ard_id="lm02", source_name=source_name)
        self.out_zip.write(resource_path("static/atlantica_skip.script"), source_name)

    def write_wardrobe_skip_assets(self):
        """Adds assets and files to the mod for wardrobe skip."""
        source_name = _relative_mod_file("wardrobe_skip.mset")
        self.mod_yml.add_asset(
            {
                "name": "obj/N_BB080_BTL.mset",
                "method": "copy",
                "source": [{"name": source_name}],
            }
        )
        self.out_zip.write(
            resource_path("static/wardrobe/N_BB080_BTL.mset"), source_name
        )

    def write_map_skip_assets(self):
        """Adds assets and files to the mod for Port Royal map skip."""
        ca_yml_source_name = _relative_mod_file("map_skip/ca.yml")
        libretto_bin_source_name = _relative_mod_file("map_skip/libretto-ca.bar")

        self.mod_yml.add_assets(
            [
                {
                    "name": "msg/us/ca.bar",
                    "multi": [
                        {"name": "msg/fr/ca.bar"},
                        {"name": "msg/gr/ca.bar"},
                        {"name": "msg/it/ca.bar"},
                        {"name": "msg/sp/ca.bar"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "ca",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [{"name": ca_yml_source_name, "language": "en"}],
                        }
                    ],
                },
                {
                    "name": "msg/jp/ca.bar",
                    "platform": "ps2",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "ca",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [{"name": ca_yml_source_name, "language": "en"}],
                        }
                    ],
                },
                {
                    "name": "msg/jp/ca.bar",
                    "platform": "pc",
                    "method": "binarc",
                    "source": [
                        {
                            "name": "ca",
                            "type": "list",
                            "method": "kh2msg",
                            "source": [
                                {
                                    "name": ca_yml_source_name,
                                    "language": "jp",  # Change this to je whenever we update the mods manager
                                }
                            ],
                        }
                    ],
                },
                {
                    "name": "libretto-ca.bar",
                    "method": "copy",
                    "source": [{"name": libretto_bin_source_name}],
                },
            ]
        )

        self.out_zip.write(resource_path("static/map_skip/ca.yml"), ca_yml_source_name)
        self.out_zip.write(
            resource_path("static/map_skip/libretto-ca.bar"), libretto_bin_source_name
        )

    def write_cmd_list_modifications(self, modified_cmd_list_yml: str):
        """Adds assets and files to the mod for command list modification."""
        source_name = _relative_mod_file(modified_cmd_list_yml)

        self.mod_yml.add_asset_source(
            "03system.bin",
            {
                "name": "cmd",
                "type": "list",
                "method": "listpatch",
                "source": [{"name": source_name, "type": "cmd"}],
            },
        )

        self.out_zip.write(
            resource_path(f"static/{modified_cmd_list_yml}"), source_name
        )

    def write_battle_level_assets(self, modified_battle_level_binary: bytearray):
        """Adds assets and files to the mod for modified battle levels."""
        source_name = _relative_mod_file("modified_btlv.bin")

        self.mod_yml.add_asset_source(
            "00battle.bin",
            {
                "method": "copy",
                "name": "btlv",
                "type": "list",
                "source": [{"name": source_name}],
            },
        )

        self.out_zip.writestr(source_name, modified_battle_level_binary)

    def write_keyblade_locking_lua(self):
        keyblade_lua_name = _relative_mod_file("keyblade_locking/F266B00B Keyblade Locking.lua")
        assets = [
            {
                "name": "scripts/F266B00B Keyblade Locking.lua",
                "method": "copy",
                "source": [{"name": keyblade_lua_name}],
            },
        ]
        self.mod_yml.add_assets(assets)

        self.out_zip.write(
            resource_path("static/keyblade_locking/keyblade.lua"), keyblade_lua_name
        )

    def write_goa_lua(self):
        goa_lua_name = _relative_mod_file("beta_files/F266B00B GoA ROM.lua")
        assets = [
            {
                "name": "scripts/F266B00B GoA ROM.lua",
                "method": "copy",
                "source": [{"name": goa_lua_name}],
            },
        ]
        self.mod_yml.add_assets(assets)

        self.out_zip.write(
            resource_path("static/F266B00B GoA ROM.lua"), goa_lua_name
        )


    def write_better_stt_assets(self, boss_enabled: bool):
        """Adds assets and files to the mod for 'Better STT'."""
        trinity_bar_source_name = _relative_mod_file("better_stt/trinity_zz.bar")
        b_ex100_mset_source_name = _relative_mod_file("better_stt/B_EX100.mset")
        f_tt010_mset_source_name = _relative_mod_file("better_stt/F_TT010.mset")
        p_ex110_mset_source_name = _relative_mod_file("better_stt/P_EX110.mset")
        # objlist_yml_source_name = _relative_mod_file("better_stt/ObjList_Better_STT.yml")
        w_ex010_rx_mset_source_name = _relative_mod_file("better_stt/W_EX010_RX.mset")
        b_ex100_sr_mset_source_name = _relative_mod_file("better_stt/B_EX100_SR.mset")

        assets = [
            {
                "name": "limit/fm/trinity_zz.bar",
                "multi": [
                    {"name": "limit/us/trinity_zz.bar"},
                    {"name": "limit/fr/trinity_zz.bar"},
                    {"name": "limit/gr/trinity_zz.bar"},
                    {"name": "limit/it/trinity_zz.bar"},
                    {"name": "limit/sp/trinity_zz.bar"},
                ],
                "method": "copy",
                "source": [{"name": trinity_bar_source_name}],
            },
            {
                "name": "obj/B_EX100.mset",
                "method": "copy",
                "source": [{"name": b_ex100_mset_source_name}],
            },
            {
                "name": "obj/F_TT010.mset",
                "method": "copy",
                "source": [{"name": f_tt010_mset_source_name}],
            },
            {
                "name": "obj/P_EX110.mset",
                "method": "copy",
                "source": [{"name": p_ex110_mset_source_name}],
            },
            # {
            #     "name": "00objentry.bin",
            #     "method": "listpatch",
            #     "type": "List",
            #     "source": [
            #         {
            #             "name": objlist_yml_source_name,
            #             "type": "objentry"
            #         }
            #     ]
            # },
            {
                "name": "obj/W_EX010_RX.mset",
                "method": "copy",
                "source": [{"name": w_ex010_rx_mset_source_name}],
            },
        ]
        if boss_enabled:
            assets.append(
                {
                    "name": "obj/B_EX100_SR.mset",
                    "method": "copy",
                    "source": [{"name": b_ex100_sr_mset_source_name}],
                }
            )
        self.mod_yml.add_assets(assets)

        self.out_zip.write(
            resource_path("static/better_stt/trinity_zz.bar"), trinity_bar_source_name
        )
        self.out_zip.write(
            resource_path("static/better_stt/B_EX100.mset"), b_ex100_mset_source_name
        )
        self.out_zip.write(
            resource_path("static/better_stt/F_TT010.mset"), f_tt010_mset_source_name
        )
        self.out_zip.write(
            resource_path("static/better_stt/P_EX110.mset"), p_ex110_mset_source_name
        )
        self.out_zip.write(
            resource_path("static/better_stt/W_EX010_RX.mset"),
            w_ex010_rx_mset_source_name,
        )
        # out_zip.write(resource_path("static/better_stt/ObjList_Better_STT.yml"), objlist_yml_source_name)
        if boss_enabled:
            self.out_zip.write(
                resource_path("static/better_stt/B_EX100_SR.mset"),
                b_ex100_sr_mset_source_name,
            )

    def write_chest_visuals_assets(
        self, chest_assignments: list[ChestVisualAssignment]
    ):
        """Adds assets and files to the mod for 'Chest Visuals Match Contents'."""

        chest_assignments_by_file: dict[str, list[ChestVisualAssignment]] = {}
        for chest_assignment in chest_assignments:
            spawn_file_path = chest_assignment.spawn_file_path
            if spawn_file_path not in chest_assignments_by_file:
                chest_assignments_by_file[spawn_file_path] = []
            chest_assignments_by_file[spawn_file_path].append(chest_assignment)

        # Ards
        for spawn_file_path, visual_assignments in chest_assignments_by_file.items():
            with open(
                resource_path(f"static/chests/ard/{spawn_file_path}.spawn")
            ) as spawn_file:
                spawn_data = yaml.safe_load(spawn_file)

            for visual_assignment in visual_assignments:
                chest_index = visual_assignment.chest_index
                spawn_data[0]["Entities"][chest_index][
                    "ObjectId"
                ] = visual_assignment.chest_visual_id

            write_yaml_to_zip_file(
                self.out_zip,
                _relative_mod_file(f"chest/ard/{spawn_file_path}.yml"),
                spawn_data,
                sort_keys=False,
            )

        self.mod_yml.add_assets(self._get_chest_visual_assets())

        # Remastered Textures
        trash_source_name = _relative_mod_file("chest/remastered/trash.dds")
        abilities_source_name = _relative_mod_file("chest/remastered/abilities.dds")
        equipment_source_name = _relative_mod_file("chest/remastered/equipment.dds")
        forms_source_name = _relative_mod_file("chest/remastered/forms.dds")
        magic_source_name = _relative_mod_file("chest/remastered/magic.dds")
        pages_source_name = _relative_mod_file("chest/remastered/pages.dds")
        reports_source_name = _relative_mod_file("chest/remastered/reports.dds")
        stats_source_name = _relative_mod_file("chest/remastered/stats.dds")
        summons_source_name = _relative_mod_file("chest/remastered/summons.dds")
        unlocks_source_name = _relative_mod_file("chest/remastered/unlocks.dds")
        weapons_source_name = _relative_mod_file("chest/remastered/weapons.dds")
        proofs_glow_source_name = _relative_mod_file("chest/remastered/proofs_glow.dds")
        proofs_source_name = _relative_mod_file("chest/remastered/proofs.dds")
        self.out_zip.write(
            resource_path("static/chests/remastered/trash.dds"), trash_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/abilities.dds"),
            abilities_source_name,
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/forms.dds"), forms_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/magic.dds"), magic_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/pages.dds"), pages_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/reports.dds"), reports_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/stats.dds"), stats_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/summons.dds"), summons_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/unlocks.dds"), unlocks_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/weapons.dds"), weapons_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/proofs_glow.dds"),
            proofs_glow_source_name,
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/proofs.dds"), proofs_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/remastered/equipment.dds"), equipment_source_name
        )

        # Listpatch
        chest_obj_list_source_name = _relative_mod_file("chest/ChestObjList.yml")
        self.out_zip.write(
            resource_path("static/chests/ChestObjList.script"),
            chest_obj_list_source_name,
        )

        # MSET
        f_ex030_lk_source_name = _relative_mod_file("chest/obj/F_EX030_LK.mset")
        f_ex040_lk_source_name = _relative_mod_file("chest/obj/F_EX040_LK.mset")
        f_ex040_rx_source_name = _relative_mod_file("chest/obj/F_EX040_RX.mset")
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_LK.mset"), f_ex030_lk_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX040_LK.mset"), f_ex040_lk_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX040_RX.mset"), f_ex040_rx_source_name
        )

        # MDLX
        jnk_source_name = _relative_mod_file("chest/obj/F_EX030_JK2.mdlx")
        abl_source_name = _relative_mod_file("chest/obj/F_EX030_ABL.mdlx")
        eqp_source_name = _relative_mod_file("chest/obj/F_EX030_EQP.mdlx")
        frm_source_name = _relative_mod_file("chest/obj/F_EX030_FRM.mdlx")
        mag_source_name = _relative_mod_file("chest/obj/F_EX030_MAG.mdlx")
        pag_source_name = _relative_mod_file("chest/obj/F_EX030_PAG.mdlx")
        rep_source_name = _relative_mod_file("chest/obj/F_EX030_REP.mdlx")
        sta_source_name = _relative_mod_file("chest/obj/F_EX030_STA.mdlx")
        smn_source_name = _relative_mod_file("chest/obj/F_EX030_SMN.mdlx")
        vst_source_name = _relative_mod_file("chest/obj/F_EX030_VST.mdlx")
        wep_source_name = _relative_mod_file("chest/obj/F_EX030_WEP.mdlx")
        prf_source_name = _relative_mod_file("chest/obj/F_EX040_PRF.mdlx")
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_JK2.mdlx"), jnk_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_ABL.mdlx"), abl_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_EQP.mdlx"), eqp_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_FRM.mdlx"), frm_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_MAG.mdlx"), mag_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_PAG.mdlx"), pag_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_REP.mdlx"), rep_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_STA.mdlx"), sta_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_SMN.mdlx"), smn_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_VST.mdlx"), vst_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX030_WEP.mdlx"), wep_source_name
        )
        self.out_zip.write(
            resource_path("static/chests/obj/F_EX040_PRF.mdlx"), prf_source_name
        )

    def validate_and_write_mod_yml(
        self, zip_data: io.BytesIO, settings: RandomizerSettings
    ) -> io.BytesIO:
        """Performs some validation and deduplication of assets in the mod, and writes the mod.yml to the zip file."""

        mod_data = self.mod_yml.data

        # merge cmd mods if there are two
        #  first, find all the list patches we are assigning
        listpatch_file_paths: list[str] = []
        listpatch_contents: list[dict[str, Any]] = []

        for modded_file in self.mod_yml.find_assets("03system.bin"):
            for source_index, source in enumerate(modded_file["source"]):
                if source["name"] == "cmd":
                    listpatch_file_paths.append(
                        source["source"][0]["name"].replace("\\", "/")
                    )

        merged_command_list_source = _relative_mod_file("cmd_list_merged.yml")

        # if there are multiple cmd listpatches, we have to read all the contents to create new file
        if len(listpatch_file_paths) > 1:
            with ZipFile(zip_data, "r") as current_zip:
                for listpatch_file_path in listpatch_file_paths:
                    path = zipfile.Path(current_zip, at=listpatch_file_path)
                    with path.open(encoding="UTF-8") as current_listpatch_file:
                        listpatch_contents.extend(
                            yaml.safe_load(current_listpatch_file)
                        )

            # remove all instances of 03_system cmd listpatches (we'll add another afterward)
            num_03system_entries = 0
            delete_asset_indices: list[int] = []
            for asset_index, asset in enumerate(mod_data["assets"]):
                delete_source_indices: list[int] = []
                if asset["name"] == "03system.bin":
                    num_03system_entries += 1
                    for source_index, source in enumerate(asset["source"]):
                        if source["name"] == "cmd":
                            delete_source_indices.append(source_index)
                    asset["source"] = [
                        src
                        for src_index, src in enumerate(asset["source"])
                        if src_index not in delete_source_indices
                    ]
                    if len(asset["source"]) == 0:
                        delete_asset_indices.append(asset_index)
            # assumes that empty 03system lists are only happening from
            # listpatches of cmd, and we don't want to remove the only
            # instance of this in the yaml, so only remove empty entries
            # if they are duplicates
            if num_03system_entries > 1:
                mod_data["assets"] = [
                    asset
                    for a_index, asset in enumerate(mod_data["assets"])
                    if a_index not in delete_asset_indices
                ]

            self.mod_yml.add_asset_source(
                "03system.bin",
                {
                    "name": "cmd",
                    "method": "listpatch",
                    "type": "list",
                    "source": [{"name": merged_command_list_source, "type": "cmd"}],
                },
            )

        for modded_file in ["00battle.bin"]:
            first_asset_index: Optional[int] = None
            delete_asset_indices: list[int] = []
            for asset_index, asset in enumerate(mod_data["assets"]):
                if asset["name"] == modded_file:
                    if first_asset_index:
                        delete_asset_indices.append(asset_index)
                        # add these sources into this group
                        mod_data["assets"][first_asset_index]["source"] += asset[
                            "source"
                        ]
                    else:
                        first_asset_index = asset_index
            # delete all duplicates
            mod_data["assets"] = [
                asset
                for a_index, asset in enumerate(mod_data["assets"])
                if a_index not in delete_asset_indices
            ]

        # add multi support for TWTNW NPCs, msg/jp/eh.bar (TODO: remove this when khbr does it on its own)
        for asset in mod_data["assets"]:
            if asset["name"] == "msg/jp/eh.bar":
                asset["multi"].append({"name": "msg/fr/eh.bar"})
                asset["multi"].append({"name": "msg/gr/eh.bar"})
                asset["multi"].append({"name": "msg/it/eh.bar"})
                asset["multi"].append({"name": "msg/sp/eh.bar"})

        # now that the mod yml is proper, we want to add any merged files into the zip, along with the mod.yml
        with ZipFile(zip_data, "a", ZIP_DEFLATED) as current_zip:
            self.mod_yml.write_to_zip_file(current_zip)
            if len(listpatch_contents) > 0:
                write_yaml_to_zip_file(
                    current_zip,
                    merged_command_list_source,
                    listpatch_contents,
                    sort_keys=False,
                )

        zip_data.seek(0)
        return zip_data

    def write_mod_ymls(self, include_main_mod_yml: bool):
        """Writes the output for the various mod YAML files."""
        zip_file = self.out_zip

        self.bonuses.write_to_zip_file(zip_file)
        self.form_levels.write_to_zip_file(zip_file)
        self.items.write_to_zip_file(zip_file)
        self.level_ups.write_to_zip_file(zip_file)
        self.messages.write_to_zip_file(zip_file)
        self.journal_txt.write_to_zip_file(zip_file)
        self.player_params.write_to_zip_file(zip_file)
        if self.prize_table.has_entries():
            # Prize table only has entries if certain settings are on, so we may need to add its asset as well
            self._add_prize_table_asset()
            self.prize_table.write_to_zip_file(zip_file)
        if self.atkp_organizer.has_entries():
            self._add_atkp_asset()
            self.atkp_organizer.write_to_zip_file(zip_file)
        self.treasures.write_to_zip_file(zip_file)

        # Main mod.yml comes last in case any additional assets were added
        if include_main_mod_yml:
            self.mod_yml.write_to_zip_file(zip_file)

    def _add_areadata_script_asset(self, ard_id: str, source_name: str):
        self.mod_yml.add_asset(
            {
                "name": f"ard/{ard_id}.ard",
                "multi": [
                    {"name": f"ard/fr/{ard_id}.ard"},
                    {"name": f"ard/gr/{ard_id}.ard"},
                    {"name": f"ard/it/{ard_id}.ard"},
                    {"name": f"ard/jp/{ard_id}.ard"},
                    {"name": f"ard/sp/{ard_id}.ard"},
                    {"name": f"ard/us/{ard_id}.ard"},
                ],
                "method": "binarc",
                "source": [
                    {
                        "name": "evt",
                        "type": "areadatascript",
                        "method": "areadatascript",
                        "source": [{"name": source_name}],
                    }
                ],
            }
        )

    def _add_prize_table_asset(self):
        self.mod_yml.add_asset_source(
            "00battle.bin",
            {
                "name": "przt",
                "type": "list",
                "method": "listpatch",
                "source": [{"name": self.prize_table.source_name, "type": "przt"}],
            },
        )

    @staticmethod
    def _get_chest_visual_assets() -> list[Asset]:
        assets: list[Asset] = []

        basic_types = {
            "JK2": "trash",
            "ABL": "abilities",
            "EQP": "equipment",
            "FRM": "forms",
            "MAG": "magic",
            "PAG": "pages",
            "REP": "reports",
            "STA": "stats",
            "SMN": "summons",
            "VST": "unlocks",
            "WEP": "weapons",
        }

        # ARD edits
        for room_1 in [
            "al00",
            "al01",
            "al06",
            "al07",
            "al10",
            "al11",
            "al12",
            "al13",
            "bb02",
            "bb03",
            "bb06",
            "bb07",
            "bb08",
            "bb09",
            "bb10",
            "ca00",
            "ca02",
            "ca09",
            "ca11",
            "ca12",
            "ca13",
            "ca14",
            "ca15",
            "dc01",
            "dc03",
            "eh02",
            "eh03",
            "eh04",
            "eh06",
            "eh09",
            "eh12",
            "eh17",
            "hb03",
            "hb09",
            "hb11",
            "hb12",
            "hb18",
            "hb21",
            "hb22",
            "hb23",
            "hb24",
            "he03",
            "he10",
            "he11",
            "he12",
            "he15",
            "he16",
            "he17",
            "lk00",
            "lk03",
            "lk05",
            "lk06",
            "lk07",
            "lk08",
            "lk09",
            "mu00",
            "mu02",
            "mu03",
            "mu05",
            "mu06",
            "mu11",
            "nm00",
            "nm01",
            "nm02",
            "nm04",
            "nm06",
            "nm08",
            "po01",
            "po02",
            "po03",
            "po04",
            "po05",
            "po09",
            "tr00",
            "tr01",
            "tr04",
            "tr05",
            "tr08",
            "wi00",
            "wi01",
            "wi02",
            "tt32",
            "tt33",
        ]:
            assets.append(
                {
                    "name": "ard/" + room_1 + ".ard",
                    "multi": [
                        {"name": "ard/jp/" + room_1 + ".ard"},
                        {"name": "ard/us/" + room_1 + ".ard"},
                        {"name": "ard/fr/" + room_1 + ".ard"},
                        {"name": "ard/gr/" + room_1 + ".ard"},
                        {"name": "ard/it/" + room_1 + ".ard"},
                        {"name": "ard/sp/" + room_1 + ".ard"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "m_70",
                            "type": "areadataspawn",
                            "method": "spawnpoint",
                            "source": [
                                {
                                    "name": _relative_mod_file(
                                        "chest/ard/" + room_1 + "/m_70.yml"
                                    )
                                }
                            ],
                        }
                    ],
                }
            )

        for room_2 in [
            "bb12",
            "hb05",
            "hb06",
            "hb26",
            "tt09",
            "tt10",
            "tt15",
            "tt16",
            "tt17",
            "tt22",
        ]:
            assets.append(
                {
                    "name": "ard/" + room_2 + ".ard",
                    "multi": [
                        {"name": "ard/jp/" + room_2 + ".ard"},
                        {"name": "ard/us/" + room_2 + ".ard"},
                        {"name": "ard/fr/" + room_2 + ".ard"},
                        {"name": "ard/gr/" + room_2 + ".ard"},
                        {"name": "ard/it/" + room_2 + ".ard"},
                        {"name": "ard/sp/" + room_2 + ".ard"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "m_70",
                            "type": "areadataspawn",
                            "method": "spawnpoint",
                            "source": [
                                {
                                    "name": _relative_mod_file(
                                        "chest/ard/" + room_2 + "/m_70.yml"
                                    )
                                }
                            ],
                        },
                        {
                            "name": "m_71",
                            "type": "areadataspawn",
                            "method": "spawnpoint",
                            "source": [
                                {
                                    "name": _relative_mod_file(
                                        "chest/ard/" + room_2 + "/m_71.yml"
                                    )
                                }
                            ],
                        },
                    ],
                }
            )

        for room_3 in [
            "tt07",
            "tt13",
            "tt14",
            "tt25",
            "tt26",
            "tt27",
            "tt28",
            "tt36",
            "tt37",
        ]:
            assets.append(
                {
                    "name": "ard/" + room_3 + ".ard",
                    "multi": [
                        {"name": "ard/jp/" + room_3 + ".ard"},
                        {"name": "ard/us/" + room_3 + ".ard"},
                        {"name": "ard/fr/" + room_3 + ".ard"},
                        {"name": "ard/gr/" + room_3 + ".ard"},
                        {"name": "ard/it/" + room_3 + ".ard"},
                        {"name": "ard/sp/" + room_3 + ".ard"},
                    ],
                    "method": "binarc",
                    "source": [
                        {
                            "name": "m_71",
                            "type": "areadataspawn",
                            "method": "spawnpoint",
                            "source": [
                                {
                                    "name": _relative_mod_file(
                                        "chest/ard/" + room_3 + "/m_71.yml"
                                    )
                                }
                            ],
                        }
                    ],
                }
            )

        # MDLXs
        for basic_type_suffix, _ in basic_types.items():
            assets.append(
                {
                    "name": f"obj/F_EX030_{basic_type_suffix}.mdlx",
                    "multi": [
                        {"name": f"obj/F_EX050_{basic_type_suffix}.mdlx"},
                    ],
                    "method": "copy",
                    "source": [
                        {
                            "name": _relative_mod_file(
                                f"chest/obj/F_EX030_{basic_type_suffix}.mdlx"
                            )
                        }
                    ],
                }
            )
        # Proofs
        assets.append(
            {
                "name": "obj/F_EX040_PRF.mdlx",
                "multi": [
                    {"name": "obj/F_EX060_PRF.mdlx"},
                ],
                "method": "copy",
                "source": [{"name": _relative_mod_file("chest/obj/F_EX040_PRF.mdlx")}],
            }
        )

        # chest *.a.us (PC)
        assets.append(
            {
                "name": "obj/F_EX030_JK2.a.us",
                "platform": "pc",
                "multi": [
                    {"name": "obj/F_EX030_ABL.a.us"},
                    {"name": "obj/F_EX030_EQP.a.us"},
                    {"name": "obj/F_EX030_FRM.a.us"},
                    {"name": "obj/F_EX030_MAG.a.us"},
                    {"name": "obj/F_EX030_PAG.a.us"},
                    {"name": "obj/F_EX030_REP.a.us"},
                    {"name": "obj/F_EX030_STA.a.us"},
                    {"name": "obj/F_EX030_SMN.a.us"},
                    {"name": "obj/F_EX030_VST.a.us"},
                    {"name": "obj/F_EX030_WEP.a.us"},
                    {"name": "obj/F_EX050_JK2.a.us"},
                    {"name": "obj/F_EX050_ABL.a.us"},
                    {"name": "obj/F_EX050_EQP.a.us"},
                    {"name": "obj/F_EX050_FRM.a.us"},
                    {"name": "obj/F_EX050_MAG.a.us"},
                    {"name": "obj/F_EX050_PAG.a.us"},
                    {"name": "obj/F_EX050_REP.a.us"},
                    {"name": "obj/F_EX050_STA.a.us"},
                    {"name": "obj/F_EX050_SMN.a.us"},
                    {"name": "obj/F_EX050_VST.a.us"},
                    {"name": "obj/F_EX050_WEP.a.us"},
                ],
                "method": "copy",
                "source": [
                    {
                        "name": "obj/F_EX030.a.us",
                        "type": "internal",
                    }
                ],
            }
        )
        assets.append(
            {
                "name": "obj/F_EX040_PRF.a.us",
                "platform": "pc",
                "multi": [
                    {"name": "obj/F_EX060_PRF.a.us"},
                ],
                "method": "copy",
                "source": [
                    {
                        "name": "obj/F_EX040_EH.a.us",
                        "type": "internal",
                    }
                ],
            }
        )

        # chest *.a.fm (PS2)
        assets.append(
            {
                "name": "obj/F_EX030_JK2.a.fm",
                "platform": "ps2",
                "multi": [
                    {"name": "obj/F_EX030_ABL.a.fm"},
                    {"name": "obj/F_EX030_EQP.a.fm"},
                    {"name": "obj/F_EX030_FRM.a.fm"},
                    {"name": "obj/F_EX030_MAG.a.fm"},
                    {"name": "obj/F_EX030_PAG.a.fm"},
                    {"name": "obj/F_EX030_REP.a.fm"},
                    {"name": "obj/F_EX030_STA.a.fm"},
                    {"name": "obj/F_EX030_SMN.a.fm"},
                    {"name": "obj/F_EX030_VST.a.fm"},
                    {"name": "obj/F_EX030_WEP.a.fm"},
                    {"name": "obj/F_EX050_JK2.a.fm"},
                    {"name": "obj/F_EX050_ABL.a.fm"},
                    {"name": "obj/F_EX050_EQP.a.fm"},
                    {"name": "obj/F_EX050_FRM.a.fm"},
                    {"name": "obj/F_EX050_MAG.a.fm"},
                    {"name": "obj/F_EX050_PAG.a.fm"},
                    {"name": "obj/F_EX050_REP.a.fm"},
                    {"name": "obj/F_EX050_STA.a.fm"},
                    {"name": "obj/F_EX050_SMN.a.fm"},
                    {"name": "obj/F_EX050_VST.a.fm"},
                    {"name": "obj/F_EX050_WEP.a.fm"},
                ],
                "method": "copy",
                "source": [
                    {
                        "name": "obj/F_EX030.a.fm",
                        "type": "internal",
                    }
                ],
            }
        )
        assets.append(
            {
                "name": "obj/F_EX040_PRF.a.fm",
                "platform": "ps2",
                "multi": [
                    {"name": "obj/F_EX060_PRF.a.fm"},
                ],
                "method": "copy",
                "source": [
                    {
                        "name": "obj/F_EX040_EH.a.fm",
                        "type": "internal",
                    }
                ],
            }
        )

        # MSETs
        # PL Small Chest
        assets.append(
            {
                "name": "obj/F_EX030_LK.mset",
                "method": "copy",
                "source": [{"name": _relative_mod_file("chest/obj/F_EX030_LK.mset")}],
            }
        )
        # PL Big Chest
        assets.append(
            {
                "name": "obj/F_EX040_LK.mset",
                "method": "copy",
                "source": [{"name": _relative_mod_file("chest/obj/F_EX040_LK.mset")}],
            }
        )
        # STT Big Chest (Credit to Zurph)
        assets.append(
            {
                "name": "obj/F_EX040_RX.mset",
                "method": "copy",
                "source": [{"name": _relative_mod_file("chest/obj/F_EX040_RX.mset")}],
            }
        )

        # Remastered textures (Credit to Kayya and Televo)
        for basic_type_suffix, basic_type_name in basic_types.items():
            assets.append(
                {
                    "name": f"remastered/obj/F_EX030_{basic_type_suffix}.mdlx/-0.dds",
                    "multi": [
                        {
                            "name": f"remastered/obj/F_EX050_{basic_type_suffix}.mdlx/-0.dds"
                        },
                    ],
                    "method": "copy",
                    "source": [
                        {
                            "name": _relative_mod_file(
                                f"chest/remastered/{basic_type_name}.dds"
                            )
                        }
                    ],
                }
            )
        # Proofs
        assets.append(
            {
                "name": "remastered/obj/F_EX040_PRF.mdlx/-0.dds",
                "multi": [
                    {"name": "remastered/obj/F_EX060_PRF.mdlx/-0.dds"},
                ],
                "method": "copy",
                "source": [
                    {"name": _relative_mod_file("chest/remastered/proofs_glow.dds")}
                ],
            }
        )
        assets.append(
            {
                "name": "remastered/obj/F_EX040_PRF.mdlx/-1.dds",
                "multi": [
                    {"name": "remastered/obj/F_EX060_PRF.mdlx/-1.dds"},
                ],
                "method": "copy",
                "source": [{"name": _relative_mod_file("chest/remastered/proofs.dds")}],
            }
        )

        # 00objentry edits
        assets.append(
            {
                "name": "00objentry.bin",
                "method": "listpatch",
                "type": "List",
                "source": [
                    {
                        "name": _relative_mod_file("chest/ChestObjList.yml"),
                        "type": "objentry",
                    }
                ],
            }
        )

        return assets

    def _add_atkp_asset(self):
        self.mod_yml.add_asset_source(
            "00battle.bin",
            {
                "name": "atkp",
                "type": "list",
                "method": "listpatch",
                "source": [{"name": self.atkp_organizer.source_name, "type": "atkp"}],
            },
        )

    def _get_atkp_organizer(self):
        return self.atkp_organizer


class CosmeticsModAppender:

    def __init__(self, out_zip: ZipFile, mod_yml: ModYml):
        self.out_zip = out_zip
        self.mod_yml = mod_yml

    def write_music_rando_assets(self, music_assets: list[Asset], music_replacements: dict[str, str]):
        self.mod_yml.add_assets(music_assets)
        self._write_music_replacements(music_replacements)

    def write_keyblade_rando_assets(self, keyblade_assets: list[Asset], keyblade_replacements: dict[str, str]):
        self.mod_yml.add_assets(keyblade_assets)
        self._write_keyblade_replacements(keyblade_replacements)

    def write_rando_themed_texture_assets(self):
        """Adds assets and files to the mod for 'Add Randomizer-Themed Textures'."""
        goa_computer_source_name = _relative_mod_file("textures/goa/randomizer-themed-computer.dds")
        tt_computer_source_name = _relative_mod_file("textures/tt-computer/randomizer-themed-computer-tt.dds")
        data_door_source_name = _relative_mod_file("textures/data-door/data-door-logo.dds")
        final_arena_logo_source_name = _relative_mod_file("textures/final-arena/randomizer-logo.dds")

        self.mod_yml.add_asset(
            {
                "name": "remastered/map/hb26.map/-40.dds",
                "platform": "pc",
                "method": "copy",
                "source": [{"name": goa_computer_source_name}],
            },
        )
        self.mod_yml.add_asset(
            {
                "name": "remastered/map/tt21.map/-84.dds",
                "platform": "pc",
                "method": "copy",
                "source": [{"name": tt_computer_source_name}],
            },
        )
        self.mod_yml.add_asset(
            {
                "name": "remastered/obj/F_WI550.mdlx/-2.dds",
                "platform": "pc",
                "method": "copy",
                "source": [{"name": data_door_source_name}],
            },
        )
        self.mod_yml.add_asset(
            {
                "name": "remastered/map/eh20.map/-4.dds",
                "platform": "pc",
                "multi": [
                    {"name": "remastered/map/eh20.map/-6.dds"},
                    {"name": "remastered/map/eh20.map/-9.dds"},
                ],
                "method": "copy",
                "source": [{"name": final_arena_logo_source_name}],
            },
        )

        self.out_zip.write(
            resource_path("static/textures/goa/randomizer-themed-computer.dds"),
            goa_computer_source_name,
        )
        self.out_zip.write(
            resource_path("static/textures/tt-computer/randomizer-themed-computer-tt.dds"),
            tt_computer_source_name,
        )
        self.out_zip.write(
            resource_path("static/textures/data-door/data-door-logo.dds"),
            data_door_source_name,
        )
        self.out_zip.write(
            resource_path("static/textures/final-arena/randomizer-logo.dds"),
            final_arena_logo_source_name,
        )

    def _write_keyblade_replacements(self, replacements: dict[str, str]):
        if len(replacements) > 0:
            keyblade_replacements_string = ""
            for original, replacement in replacements.items():
                keyblade_replacements_string += f"[{original}] was replaced by [{replacement}]\n"
            self.out_zip.writestr("keyblade-replacement-list.txt", keyblade_replacements_string)

    def _write_music_replacements(self, replacements: dict[str, str]):
        if len(replacements) > 0:
            music_replacements_string = ""
            for original, replacement in replacements.items():
                music_replacements_string += f"[{original}] was replaced by [{replacement}]\n"
            self.out_zip.writestr("music-replacement-list.txt", music_replacements_string)



from typing import Iterator, List, Optional

from altgraph.Graph import Graph

from Class.newLocationClass import KH2Location
from List.configDict import locationCategory
from List.location import landofdragons, spaceparanoids, weaponslot, donaldbonus, goofybonus, starting, formlevel, \
    summonlevel, agrabah, disneycastle, hundredacrewood, olympuscoliseum, beastscastle, halloweentown, portroyal, \
    hollowbastion, pridelands, simulatedtwilighttown, twilighttown, worldthatneverwas, atlantica, synthesis, \
    puzzlereward, soralevel
from List.location.graph import LocationGraphBuilder, START_NODE, RequirementEdge, RequirementFunction
from Module.RandomizerSettings import RandomizerSettings


# updated function to allow for multiple incoming connections to enforce multiple constraints
def get_all_parent_edge_requirements(
        node_id: str,
        graph: Graph
) -> RequirementFunction:
    # base_requirement = lambda inv : True
    non_strict_requirements = []
    strict_requirements = []

    for edge in graph.inc_edges(node_id):
        # we want to collate the parent requirements in the proper arrangement. 
        #   If we have non-strict edges, only one of the parent edges is needed to enter this node.
        source, _ = graph.edge_by_id(edge)
        parent_requirements = get_all_parent_edge_requirements(source, graph)
        # if debug_mode:
        #     parent_requirements([])
        data: RequirementEdge = graph.edge_data(edge)
        # condition to get to here through this parent
        if data.requirement is not None:
            requirement = lambda inv, this_req=data.requirement, par_req=parent_requirements : this_req(inv) and par_req(inv)
        else:
            requirement = lambda inv, par_req=parent_requirements: par_req(inv)
        if data.strict:
            strict_requirements.append(requirement)
        else:
            non_strict_requirements.append(requirement)
    if len(non_strict_requirements) > 0:
        #convert the non-strict requirements into a strict with with an or condition
        new_strict_func = lambda inv, non_s_req=non_strict_requirements: any([r(inv) for r in non_s_req])
        strict_requirements.append(new_strict_func)
    this_requirement = lambda inv, s_req=strict_requirements : all([r(inv) for r in s_req])
    return this_requirement


class Locations:

    def __init__(self, settings: RandomizerSettings, secondary_graph: bool = False):
        self.location_graph = Graph()
        self.reverse_rando = secondary_graph
        self.locations_by_name: dict[str, KH2Location] = {}
        self.first_boss_nodes: list[str] = []
        self.last_story_boss_nodes: list[str] = []
        self.superboss_nodes: list[str] = []
        self.make_location_graph(settings)

    def _all_locations_iter(self) -> Iterator[KH2Location]:
        graph = self.location_graph
        for node_id in graph.nodes.keys():
            node_data = graph.node_data(node_id)
            for location in node_data.locations:
                yield location

    def _common_ancestor(self, node_id) -> List[KH2Location]:
        result = [node_id]
        current_node = node_id
        graph = self.location_graph

        while True:
            in_nbrs = graph.inc_nbrs(current_node)
            assert len(in_nbrs)==1
            current_node = in_nbrs[0]
            result.append(current_node)
            if current_node==START_NODE:
                break
            # check all out edges
            bfs_search = [current_node]
            while len(bfs_search):
                bfs_node = bfs_search.pop(0)
                out_nbrs = graph.out_nbrs(bfs_node)
                for o in out_nbrs:
                    if o not in result:
                        bfs_search.append(o)
                        result.append(o)

        return result

    def node_ids(self) -> list[str]:
        """ Returns a list of all the node IDs in the graph. """
        return self.location_graph.node_list()

    def node_ids_before(self, node_id: str, include_self: bool = False, include_starting: bool = False) -> list[str]:
        """
        Returns a list of all node IDs that come before the given node_id. Additional argument(s) can be given to
        control whether the given node_id itself is included in the returned list, as well as the starting node.
        """
        result = []
        ancestors = self._common_ancestor(node_id)
        for before_node_id in ancestors:
            if node_id == before_node_id and not include_self:
                continue
            if before_node_id == START_NODE and not include_starting:
                continue
            result.append(before_node_id)
        return result

    def node_ids_after(self, node_id: str, include_self: bool = False) -> list[str]:
        """
        Returns a list of all node IDs that come after the given node_id. An additional argument can be given to
        control whether the given node_id itself is included in the returned list.
        """
        result = []
        for after_node_id in self.location_graph.forw_bfs(node_id):
            if node_id == after_node_id and not include_self:
                continue
            result.append(after_node_id)
        return result

    def locations_for_node(self, node_id: str) -> list[KH2Location]:
        """ Returns a list of the locations for the given node ID. """
        return self.location_graph.node_data(node_id).locations

    def locations_for_category(self, category: locationCategory) -> list[KH2Location]:
        """ Returns all locations whose category matches the input category. """
        return [location for location in self._all_locations_iter() if category == location.LocationCategory]

    def all_locations(self) -> list[KH2Location]:
        """ Returns a list of all the locations in the graph. """
        result = []
        for node_id in self.node_ids():
            result.extend(self.locations_for_node(node_id))
        return result

    def locations_before(
            self,
            node_id: str,
            include_self: bool = False,
            include_starting_node: bool = False
    ) -> list[KH2Location]:
        """
        Returns a list of all locations that come before the given node_id. Additional argument(s) can be given to
        control whether locations for the given node_id itself are included in the returned list, as well as the
        starting node.
        """
        result = []
        for before_node_id in self.node_ids_before(node_id, include_self, include_starting_node):
            result.extend(self.locations_for_node(before_node_id))
        return result

    def locations_after(self, node_id: str, include_self: bool = False) -> list[KH2Location]:
        """
        Returns a list of all locations that come after the given node_id. An additional argument can be given to
        control whether locations for the given node_id itself are included in the returned list.
        """
        result = []
        for after_node_id in self.node_ids_after(node_id, include_self):
            result.extend(self.locations_for_node(after_node_id))
        return result

    @staticmethod
    def all_donald_locations() -> list[KH2Location]:
        return donaldbonus.donald_bonuses() + weaponslot.donald_staff_slots() + starting.donald_starting_items()

    @staticmethod
    def all_goofy_locations() -> list[KH2Location]:
        return goofybonus.goofy_bonuses() + weaponslot.goofy_shield_slots() + starting.goofy_starting_items()

    def make_location_graph(self, settings: RandomizerSettings):
        builder = LocationGraphBuilder(self.location_graph, self.reverse_rando, settings)

        starting.make_graph(builder)
        simulatedtwilighttown.make_graph(builder)
        twilighttown.make_graph(builder)
        hollowbastion.make_graph(builder)
        landofdragons.make_graph(builder)
        beastscastle.make_graph(builder)
        olympuscoliseum.make_graph(builder)
        disneycastle.make_graph(builder)
        portroyal.make_graph(builder)
        agrabah.make_graph(builder)
        halloweentown.make_graph(builder)
        pridelands.make_graph(builder)
        spaceparanoids.make_graph(builder)
        worldthatneverwas.make_graph(builder)
        hundredacrewood.make_graph(builder)
        atlantica.make_graph(builder)
        soralevel.make_graph(builder)
        formlevel.make_graph(builder)
        summonlevel.make_graph(builder)
        synthesis.make_graph(builder)
        puzzlereward.make_graph(builder)

        for name, location in builder.pending_locations_by_name.items():
            self.locations_by_name[name] = location
        self.first_boss_nodes.extend(builder.pending_first_boss_nodes)
        self.last_story_boss_nodes.extend(builder.pending_last_story_boss_nodes)
        self.superboss_nodes.extend(builder.pending_superboss_nodes)

        # from altgraph.Dot import Dot
        # dot = Dot(self.location_graph)
        # dot.style(rankdir="LR")
        # if self.reverse_rando:
        #     dot.save_img(file_name='reverse_graph',file_type="gif")
        # else:
        #     dot.save_img(file_name='graph',file_type="gif")



JUST_DAMAGE = "Just Damage"
DAMAGE_STUN = "Stun"
DAMAGE_STUN_KNOCKBACK = "Stun/Knockback"


class KnockbackTypes:

    def __init__(self, knockback_choice: str):
        super().__init__()
        self.knockback_choice = knockback_choice

    @staticmethod
    def knockback_options() -> dict[str, str]:
        return {
            JUST_DAMAGE: "Just Damage",
            DAMAGE_STUN: "Stun",
            DAMAGE_STUN_KNOCKBACK: "Knockback",
        }

    @staticmethod
    def get_knockback_value(knockback) -> int:
        if knockback == DAMAGE_STUN:
            return 8
        elif knockback == DAMAGE_STUN_KNOCKBACK:
            return 11
        else:
            return 12

    @staticmethod
    def damage_type_tooltip() -> str:
        return """
        Defines the type of knockback the attack/ability will have.
        Keep in mind, changes will also affect enemies that Companions could already affect normally.
        
        Just Damage: Enemies won't be phased (except for certain things like Goofy Tornado which have
        a second attack effect for pulling enemies in).
        
        Damage/Stun: Enemies will be stunned, but won't move very much or at all from their position.
        
        Damage/Stun/Knockback: Like above, but will also be moved.
        """



from dataclasses import dataclass
from enum import Enum

class ObjectiveType(str, Enum):
    BOSS = "Boss"
    WORLDPROGRESS = "WorldProgress"
    FIGHT = "Fight"

class ObjectiveDifficulty(str, Enum):
    EARLY = "Early"
    MIDDLE = "Middle"
    LATE = "Late"
    LATEST = "Latest"

@dataclass (frozen=True)
class KH2Objective:
    Name: str
    Location: Enum
    Type: ObjectiveType
    Difficulty: ObjectiveDifficulty = ObjectiveDifficulty.EARLY

def get_last_story_objectives():
    from List.location import (simulatedtwilighttown,twilighttown,hollowbastion,landofdragons,
                            beastscastle,disneycastle,portroyal,agrabah,halloweentown,
                            pridelands, spaceparanoids, worldthatneverwas, olympuscoliseum,
                            hundredacrewood, atlantica, formlevel,puzzlereward)
    return [
        KH2Objective("Defeat Axel II",simulatedtwilighttown.CheckLocation.Axel2,ObjectiveType.BOSS),
        KH2Objective("Fight alongside Axel",twilighttown.CheckLocation.BetwixtAndBetween,ObjectiveType.FIGHT,ObjectiveDifficulty.LATE),
        KH2Objective("Stop the 1000 Heartless",hollowbastion.CheckLocation.ThousandHeartless,ObjectiveType.FIGHT,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Stormrider",landofdragons.CheckLocation.StormRiderBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Xaldin",beastscastle.CheckLocation.XaldinBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Future Pete",disneycastle.CheckLocation.FuturePete,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Grim Reaper II",portroyal.CheckLocation.GrimReaper2,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Genie Jafar",agrabah.CheckLocation.GenieJafarBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat the Experiment",halloweentown.CheckLocation.Experiment,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Groundshaker",pridelands.CheckLocation.Groundshaker,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat MCP",spaceparanoids.CheckLocation.McpBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Xemnas",worldthatneverwas.CheckLocation.Xemnas1Bonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Defeat Hades",olympuscoliseum.CheckLocation.Hades,ObjectiveType.BOSS,ObjectiveDifficulty.LATE),
        KH2Objective("Help Pooh out of the pot of hunny",hundredacrewood.CheckLocation.StarryHillCureElement,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        KH2Objective("Participate the New Day Musical",atlantica.CheckLocation.MusicalBlizzardElement,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
    ]

def get_full_objective_list():
    from List.location import (simulatedtwilighttown,twilighttown,hollowbastion,landofdragons,
                            beastscastle,disneycastle,portroyal,agrabah,halloweentown,
                            pridelands, spaceparanoids, worldthatneverwas, olympuscoliseum,
                            hundredacrewood, atlantica, formlevel,puzzlereward)
    return get_last_story_objectives() + [
        # STT
        KH2Objective("Defeat Twilight Thorn",simulatedtwilighttown.CheckLocation.TwilightThorn,ObjectiveType.BOSS),
        KH2Objective("Defeat Axel I",simulatedtwilighttown.CheckLocation.Axel1,ObjectiveType.FIGHT),
        KH2Objective("Fight Setzer",simulatedtwilighttown.CheckLocation.StruggleTrophy,ObjectiveType.FIGHT),
        # TT
        KH2Objective("Talk to the 3 Fairies",twilighttown.CheckLocation.ValorForm,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Sandlot Berserkers",twilighttown.CheckLocation.SeifersTrophy,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        # HB
        KH2Objective("Defend the Bailey",hollowbastion.CheckLocation.Bailey,ObjectiveType.FIGHT),
        KH2Objective("Defeat Demyx",hollowbastion.CheckLocation.DemyxHollowBastion,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # CoR
        KH2Objective("Reach the end of Cavern of Remembrance",hollowbastion.CheckLocation.CorMineshaftUpperLevelApBoost,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        KH2Objective("Reach the end of Transport of Remembrance",hollowbastion.CheckLocation.TransportToRemembrance,ObjectiveType.FIGHT,ObjectiveDifficulty.LATEST),
        # LoD
        KH2Objective("Climb the Mountain Trail",landofdragons.CheckLocation.VillageCaveAreaMap,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Fight the Enemies in Village Cave",landofdragons.CheckLocation.VillageCaveBonus,ObjectiveType.FIGHT),
        KH2Objective("Defeat Shan Yu",landofdragons.CheckLocation.ShanYuBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # BC
        KH2Objective("Defeat Thresholder",beastscastle.CheckLocation.Thresholder,ObjectiveType.BOSS),
        KH2Objective("Help Beast",beastscastle.CheckLocation.Beast,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Dark Thorn",beastscastle.CheckLocation.DarkThorn,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # DC
        KH2Objective("Escort Queen Minnie",disneycastle.CheckLocation.MinnieEscort,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Fight through Windows of Time",disneycastle.CheckLocation.WindowOfTimeMap,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Stop Pete from Escaping",disneycastle.CheckLocation.BoatPete,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        # PR
        KH2Objective("Stall for time on Isla de Muerta",portroyal.CheckLocation.IslaDeMuertaMap,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defend the Interceptor from pirates",portroyal.CheckLocation.BoatFight,ObjectiveType.FIGHT),
        KH2Objective("Stop the explosive barrels",portroyal.CheckLocation.InterceptorBarrels,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Barbossa",portroyal.CheckLocation.Barbossa,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat Grim Reaper I",portroyal.CheckLocation.GrimReaper1,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # AG
        KH2Objective("Escort Abu",agrabah.CheckLocation.AbuEscort,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Survive the Treasure Room Ambush",agrabah.CheckLocation.TreasureRoomBonus,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat the Elemental Lords",agrabah.CheckLocation.ElementalLordsBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # HT
        KH2Objective("Defeat Prison Keeper",halloweentown.CheckLocation.PrisonKeeper,ObjectiveType.BOSS),
        KH2Objective("Defeat Oogie Boogie",halloweentown.CheckLocation.OogieBoogie,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Capture Lock Shock and Barrel",halloweentown.CheckLocation.LockShockBarrel,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Find the stolen presents",halloweentown.CheckLocation.Present,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Create decoy presents",halloweentown.CheckLocation.DecoyPresents,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        # PL
        KH2Objective("Reunite with Simba",pridelands.CheckLocation.CircleOfLife,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Rescue Timon and Pumbaa",pridelands.CheckLocation.Hyenas1,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Scar",pridelands.CheckLocation.Scar,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Get info about the ghost of Scar from Hyenas",pridelands.CheckLocation.Hyenas2,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        # SP
        KH2Objective("Survive the Dataspace Attack",spaceparanoids.CheckLocation.ScreensBonus,ObjectiveType.FIGHT),
        KH2Objective("Defeat Hostile Program",spaceparanoids.CheckLocation.HostileProgramBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Ride the Solar Sailor",spaceparanoids.CheckLocation.SolarSailerBonus,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        # TWTNW
        KH2Objective("Defeat Roxas",worldthatneverwas.CheckLocation.Roxas,ObjectiveType.BOSS),
        KH2Objective("Defeat Xigbar",worldthatneverwas.CheckLocation.XigbarBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat Luxord",worldthatneverwas.CheckLocation.LuxordBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat Saix",worldthatneverwas.CheckLocation.SaixBonus,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        # OC
        KH2Objective("Defeat Cerberus",olympuscoliseum.CheckLocation.Cerberus,ObjectiveType.BOSS),
        KH2Objective("Train with Phil",olympuscoliseum.CheckLocation.Urns,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Pete at The Lock",olympuscoliseum.CheckLocation.PeteOlympusColiseum,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat Hydra",olympuscoliseum.CheckLocation.Hydra,ObjectiveType.BOSS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Defeat the Ambush in Hades Chamber",olympuscoliseum.CheckLocation.AuronssStatue,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        # Cups
        KH2Objective("Win the Pain and Panic Cup",olympuscoliseum.CheckLocation.PainPanicCupProtectBelt,ObjectiveType.FIGHT),
        KH2Objective("Win the Cerberus Cup",olympuscoliseum.CheckLocation.CerberusCupRisingDragon,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Win the Titan Cup",olympuscoliseum.CheckLocation.TitanCupGenjiShield,ObjectiveType.FIGHT,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Win the Goddess of Fate Cup",olympuscoliseum.CheckLocation.GoddessOfFateCupFatalCrest,ObjectiveType.FIGHT,ObjectiveDifficulty.LATE),
        # HAW
        KH2Objective("Rescue Pooh from the Spooky Cave",hundredacrewood.CheckLocation.SweetMemories,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        # AT
        KH2Objective("Relearn to swim",atlantica.CheckLocation.UnderseaKingdomMap,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Defeat Ursula",atlantica.CheckLocation.MysteriousAbyss,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        # FORM
        KH2Objective("Reach Valor Level 3",formlevel.CheckLocation.Valor3,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Reach Valor Level 5",formlevel.CheckLocation.Valor5,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Reach Valor Level 7",formlevel.CheckLocation.Valor7,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        KH2Objective("Reach Wisdom Level 3",formlevel.CheckLocation.Wisdom3,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Reach Wisdom Level 5",formlevel.CheckLocation.Wisdom5,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Reach Wisdom Level 7",formlevel.CheckLocation.Wisdom7,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        KH2Objective("Reach Limit Level 3",formlevel.CheckLocation.Limit3,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Reach Limit Level 5",formlevel.CheckLocation.Limit5,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Reach Limit Level 7",formlevel.CheckLocation.Limit7,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        KH2Objective("Reach Master Level 3",formlevel.CheckLocation.Master3,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Reach Master Level 5",formlevel.CheckLocation.Master5,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Reach Master Level 7",formlevel.CheckLocation.Master7,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        KH2Objective("Reach Final Level 3",formlevel.CheckLocation.Final3,ObjectiveType.WORLDPROGRESS),
        KH2Objective("Reach Final Level 5",formlevel.CheckLocation.Final5,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Reach Final Level 7",formlevel.CheckLocation.Final7,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATEST),
        # Puzzle
        KH2Objective("Complete the Awakening Puzzle",puzzlereward.CheckLocation.AwakeningApBoost,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Complete the Heart Puzzle",puzzlereward.CheckLocation.HeartSerenityCrystal,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.MIDDLE),
        KH2Objective("Complete the Duality Puzzle",puzzlereward.CheckLocation.DualityRareDocument,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        KH2Objective("Complete the Frontier Puzzle",puzzlereward.CheckLocation.FrontierManifestIllusion,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        KH2Objective("Complete the Daylight Puzzle",puzzlereward.CheckLocation.DaylightExecutivesRing,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        KH2Objective("Complete the Sunset Puzzle",puzzlereward.CheckLocation.SunsetGrandRibbon,ObjectiveType.WORLDPROGRESS,ObjectiveDifficulty.LATE),
        # Superbosses
        KH2Objective("Defeat Sephiroth",hollowbastion.CheckLocation.SephirothBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Lingering Will",disneycastle.CheckLocation.LingeringWillBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Roxas",simulatedtwilighttown.CheckLocation.DataRoxasMagicBoost,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Demyx",hollowbastion.CheckLocation.DataDemyxApBoost,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Xigbar",landofdragons.CheckLocation.DataXigbarDefenseBoost,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Saix",pridelands.CheckLocation.DataSaix,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Axel",twilighttown.CheckLocation.DataAxelMagicBoost,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Xaldin",beastscastle.CheckLocation.DataXaldin,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Luxord",portroyal.CheckLocation.DataLuxordApBoost,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Xemnas",worldthatneverwas.CheckLocation.DataXemnas,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Zexion",olympuscoliseum.CheckLocation.DataZexionLostIllusion,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Vexen",halloweentown.CheckLocation.DataVexen,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Larxene",spaceparanoids.CheckLocation.DataLarxeneLostIllusion,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Lexaeus",agrabah.CheckLocation.DataLexaeus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat Data Marluxia",disneycastle.CheckLocation.DataMarluxiaLostIllusion,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        # AS
        KH2Objective("Defeat AS Zexion",olympuscoliseum.CheckLocation.ZexionBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat AS Vexen",halloweentown.CheckLocation.VexenBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat AS Larxene",spaceparanoids.CheckLocation.LarxeneBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat AS Lexaeus",agrabah.CheckLocation.LexaeusBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        KH2Objective("Defeat AS Marluxia",disneycastle.CheckLocation.MarluxiaBonus,ObjectiveType.BOSS,ObjectiveDifficulty.LATEST),
        
    ]